"""
Coupled or non-coupled 4x4 linear motion
"""

from __future__ import annotations

import warnings
from collections.abc import Callable
from math import sqrt, pi, sin, cos, atan2

import numpy as np
from scipy.linalg import solve

from .amat import a_matrix, jmat, jmatswap
from .harmonic_analysis import get_tunes_harmonic
from .matrix import find_m44, find_m66
from .orbit import find_orbit4, find_orbit6
from ..constants import clight
from ..lattice import AtError, AtWarning, RFCavity
from ..lattice import Lattice, Orbit, Refpts, check_6d, get_s_pos
from ..lattice import DConstant, get_bool_index, get_uint32_index
from ..lattice import frequency_control, set_value_refpts, get_value_refpts
from ..tracking import internal_lpass

__all__ = [
    "linopt",
    "linopt2",
    "linopt4",
    "linopt6",
    "avlinopt",
    "get_optics",
    "get_tune",
    "get_chrom",
]

_S = jmat(1)
_S2 = np.array([[0, 1], [-1, 0]], dtype=np.float64)

# dtype for structured array containing linopt parameters
_DATA2_DTYPE = [
    ("alpha", np.float64, (2,)),
    ("beta", np.float64, (2,)),
    ("mu", np.float64, (2,)),
]

_DATA4_DTYPE = [
    ("alpha", np.float64, (2,)),
    ("beta", np.float64, (2,)),
    ("mu", np.float64, (2,)),
    ("gamma", np.float64),
    ("A", np.float64, (2, 2)),
    ("B", np.float64, (2, 2)),
    ("C", np.float64, (2, 2)),
]

_DATA6_DTYPE = [
    ("alpha", np.float64, (2,)),
    ("beta", np.float64, (2,)),
    ("mu", np.float64, (3,)),
    ("R", np.float64, (3, 6, 6)),
    ("A", np.float64, (6, 6)),
    ("dispersion", np.float64, (4,)),
]

_DATAX_DTYPE = [
    ("alpha", np.float64, (2,)),
    ("beta", np.float64, (2,)),
    ("mu", np.float64, (2,)),
    ("R", np.float64, (2, 4, 4)),
    ("A", np.float64, (4, 4)),
]

_W24_DTYPE = [
    ("W", np.float64, (2,)),
    ("Wp", np.float64, (2,)),
    ("dalpha", np.float64, (2,)),
    ("dbeta", np.float64, (2,)),
    ("dmu", np.float64, (2,)),
    ("ddispersion", np.float64, (4,)),
]


_W6_DTYPE = [
    ("W", np.float64, (2,)),
    ("Wp", np.float64, (2,)),
    ("dalpha", np.float64, (2,)),
    ("dbeta", np.float64, (2,)),
    ("dmu", np.float64, (3,)),
    ("dR", np.float64, (3, 6, 6)),
    ("ddispersion", np.float64, (4,)),
]

_WX_DTYPE = [
    ("W", np.float64, (2,)),
    ("Wp", np.float64, (2,)),
    ("dalpha", np.float64, (2,)),
    ("dbeta", np.float64, (2,)),
    ("dmu", np.float64, (2,)),
    ("dR", np.float64, (2, 4, 4)),
    ("ddispersion", np.float64, (4,)),
]


_IDX_DTYPE = [("idx", np.uint32)]
warnings.filterwarnings("always", category=AtWarning, module=__name__)


def _twiss22(t12, alpha0, beta0):
    """Propagate Twiss parameters"""
    bbb = t12[:, 0, 1]
    aaa = t12[:, 0, 0] * beta0 - bbb * alpha0
    beta = (aaa * aaa + bbb * bbb) / beta0
    alpha = (
        -(aaa * (t12[:, 1, 0] * beta0 - t12[:, 1, 1] * alpha0) + bbb * t12[:, 1, 1])
        / beta0
    )
    mu = np.arctan2(bbb, aaa)
    # Unwrap negative jumps in betatron phase advance
    # dmu = np.diff(np.append([0], mu))
    # jumps = dmu < -1.0e-3
    # mu += np.cumsum(jumps) * 2.0 * np.pi
    return alpha, beta, mu


def _closure(m22):
    diff = (m22[0, 0] - m22[1, 1]) / 2.0
    try:
        sinmu = np.sign(m22[0, 1]) * sqrt(-m22[0, 1] * m22[1, 0] - diff * diff)
        cosmu = 0.5 * np.trace(m22)
        alpha = diff / sinmu
        beta = m22[0, 1] / sinmu
        return alpha, beta, cosmu + sinmu * 1j
    except ValueError:  # Unstable motion
        return np.nan, np.nan, np.nan


# noinspection PyShadowingNames,PyPep8Naming
def _tunes(ring, **kwargs):
    """"""
    if ring.is_6d:
        nd = 3
        mt, _ = find_m66(ring, **kwargs)
    else:
        nd = 2
        mt, _ = find_m44(ring, **kwargs)
    try:
        _, vps = a_matrix(mt)
        tunes = np.mod(np.angle(vps) / 2.0 / pi, 1.0)
    except AtError:
        warnings.warn(AtWarning("Unstable ring"), stacklevel=1)
        tunes = np.empty(nd)
        tunes[:] = np.nan
    return tunes


def _analyze2(mt, ms):
    """Analysis of a 2D 1-turn transfer matrix"""
    A = mt[:2, :2]
    B = mt[2:, 2:]
    alp0_a, bet0_a, vp_a = _closure(A)
    alp0_b, bet0_b, vp_b = _closure(B)
    vps = np.array([vp_a, vp_b])
    el0 = (np.array([alp0_a, alp0_b]), np.array([bet0_a, bet0_b]), 0.0)
    alpha_a, beta_a, mu_a = _twiss22(ms[:, :2, :2], alp0_a, bet0_a)
    alpha_b, beta_b, mu_b = _twiss22(ms[:, 2:, 2:], alp0_b, bet0_b)
    els = (
        np.stack((alpha_a, alpha_b), axis=1),
        np.stack((beta_a, beta_b), axis=1),
        np.stack((mu_a, mu_b), axis=1),
    )
    return vps, _DATA2_DTYPE, el0, els, _W24_DTYPE


def _analyze4(mt, ms):
    """Analysis of a 4D 1-turn transfer matrix according to Sagan, Rubin"""

    def propagate(t12):
        M = t12[:2, :2]
        N = t12[2:, 2:]
        m = t12[:2, 2:]
        n = t12[2:, :2]
        ff = n @ C + g * N
        gamma = sqrt(np.linalg.det(ff))
        e12 = (g * M - m @ _S @ C.T @ _S.T) / gamma
        f12 = ff / gamma
        a12 = e12 @ A @ _S @ e12.T @ _S.T
        b12 = f12 @ B @ _S @ f12.T @ _S.T
        c12 = M @ C + g * m @ _S @ f12.T @ _S.T
        return e12, f12, gamma, a12, b12, c12

    M = mt[:2, :2]
    N = mt[2:, 2:]
    m = mt[:2, 2:]
    n = mt[2:, :2]
    H = m + _S @ n.T @ _S.T
    detH = np.linalg.det(H)
    if detH == 0.0:
        g = 1.0
        C = -H
        A = M
        B = N
    else:
        t = np.trace(M - N)
        t2 = t * t
        t2h = t2 + 4.0 * detH
        g2 = (1.0 + sqrt(t2 / t2h)) / 2
        g = sqrt(g2)
        C = -H * np.sign(t) / (g * sqrt(t2h))
        A = g2 * M - g * (m @ _S @ C.T @ _S.T + C @ n) + C @ N @ _S @ C.T @ _S.T
        B = g2 * N + g * (_S @ C.T @ _S.T @ m + n @ C) + _S @ C.T @ _S.T @ M @ C
    alp0_a, bet0_a, vp_a = _closure(A)
    alp0_b, bet0_b, vp_b = _closure(B)
    vps = np.array([vp_a, vp_b])
    el0 = (np.array([alp0_a, alp0_b]), np.array([bet0_a, bet0_b]), 0.0, g, A, B, C)
    if ms.shape[0] > 0:
        e, f, g, ai, bi, ci = zip(*[propagate(mi) for mi in ms])
        alp_a, bet_a, mu_a = _twiss22(np.array(e), alp0_a, bet0_a)
        alp_b, bet_b, mu_b = _twiss22(np.array(f), alp0_b, bet0_b)
        els = (
            np.stack((alp_a, alp_b), axis=1),
            np.stack((bet_a, bet_b), axis=1),
            np.stack((mu_a, mu_b), axis=1),
            np.array(g),
            np.stack(ai, axis=0),
            np.stack(bi, axis=0),
            np.stack(ci, axis=0),
        )
    else:
        els = (
            np.empty((0, 2)),
            np.empty((0, 2)),
            np.empty((0, 2)),
            np.empty((0,)),
            np.empty((0, 2, 2)),
            np.empty((0, 2, 2)),
            np.empty((0, 2, 2)),
        )
    return vps, _DATA4_DTYPE, el0, els, _W24_DTYPE


def _analyze6(mt, ms):
    """Analysis of a 2D, 4D, 6D 1-turn transfer matrix
    according to Wolski"""

    def get_phase(a22):
        """Return the phase for A standardization"""
        return atan2(a22[0, 1], a22[0, 0])

    def standardize(aa, slcs):
        """Apply rotation to put A in std form"""

        def rot2(slc):
            rot = -get_phase(aa[slc, slc])
            cs = cos(rot)
            sn = sin(rot)
            return aa[:, slc] @ np.array([[cs, sn], [-sn, cs]])

        return np.concatenate([rot2(slc) for slc in slcs], axis=1)

    def r_matrices(ai):
        # Rk = A * S * Ik * inv(A) * S.T
        def mul2(slc):
            return ai[:, slc] @ tt[slc, slc]

        ais = np.concatenate([mul2(slc) for slc in slices], axis=1)
        invai = solve(ai, ss.T)
        ri = np.array([ais[:, sl] @ invai[sl, :] for sl in slices])
        mui = np.array([get_phase(ai[sl, sl]) for sl in slices])
        return mui, ri, ai

    def propagate4(ri, phi, ai):
        betai = np.stack((ri[..., 0, 0, 0], ri[..., 1, 2, 2]), axis=-1)
        alphai = -np.stack((ri[..., 0, 1, 0], ri[..., 1, 3, 2]), axis=-1)
        return alphai, betai, phi, ri, ai

    def propagate6(ri, phi, ai):
        alphai, betai, phi, ri, ai = propagate4(ri, phi, ai)
        dispersion = ri[..., 2, :4, 4] / ri[..., 2, 4, 4, np.newaxis]
        return alphai, betai, phi, ri, ai, dispersion

    nv = mt.shape[0]
    dms = nv // 2
    slices = [slice(2 * i, 2 * (i + 1)) for i in range(dms)]
    ss = jmat(dms)
    tt = jmatswap(dms)
    if dms >= 3:
        propagate = propagate6
        dtype = _DATA6_DTYPE
        wtype = _W6_DTYPE
    else:
        propagate = propagate4
        dtype = _DATAX_DTYPE
        wtype = _WX_DTYPE
    a0, vps = a_matrix(mt)

    astd = standardize(a0, slices)
    phi0, r0, _ = r_matrices(astd)
    el0 = propagate(r0, phi0, astd)
    if ms.shape[0] > 0:
        ps, rs, aas = zip(*[r_matrices(mi @ astd) for mi in ms])
        els = propagate(np.array(rs), np.array(ps), np.array(aas))
    elif dms >= 3:
        els = (
            np.empty((0, dms)),
            np.empty((0, dms)),
            np.empty((0, dms)),
            np.empty((0, dms, nv, nv)),
            np.empty((0, nv, nv)),
            np.empty((0, 4)),
        )
    else:
        els = (
            np.empty((0, dms)),
            np.empty((0, dms)),
            np.empty((0, dms)),
            np.empty((0, dms, nv, nv)),
            np.empty((0, nv, nv)),
        )
    return vps, dtype, el0, els, wtype


# noinspection PyShadowingNames,PyPep8Naming
def _linopt(
    ring: Lattice,
    analyze,
    refpts=None,
    dp=None,
    dct=None,
    df=None,
    orbit=None,
    twiss_in=None,
    get_chrom=False,
    get_w=False,
    keep_lattice=False,
    mname="M",
    add0=(),
    adds=(),
    cavpts=None,
    **kwargs,
):
    """"""

    def build_sigma(orbit, dp=None):
        """Build the initial distribution at entrance of the transfer line"""
        try:
            d0 = twiss_in["dispersion"]
        except (ValueError, KeyError):  # record arrays throw ValueError !
            d0 = np.zeros((4,))

        try:
            rmat = twiss_in["R"]
        except (ValueError, KeyError):  # record arrays throw ValueError !
            rmat = None

        try:
            alphas = twiss_in["alpha"]
            betas = twiss_in["beta"]
        except (ValueError, KeyError):  # record arrays throw ValueError !
            alphas = np.zeros((2,))
            betas = np.ones((2,))

        dorbit = np.hstack((d0, 1.0, 0.0))
        if orbit is None:
            try:
                orbit = twiss_in["closed_orbit"]
            except (ValueError, KeyError):  # record arrays throw ValueError !
                orbit = np.zeros((6,))
            if dp is not None:
                orbit = orbit + dorbit * dp

        if dp is not None:
            try:
                drmat = twiss_in["dR"]
            except (ValueError, KeyError):  # record arrays throw ValueError !
                drmat = None

            try:
                dd0 = twiss_in["ddispersion"]
                dalpha = twiss_in["dalpha"]
                dbeta = twiss_in["dbeta"]
            except (ValueError, KeyError) as exc:  # record arrays throw ValueError !
                msg = (
                    "'get_w' option for a line requires 'twiss_in' calculated "
                    "with 'get_w' activated"
                )
                raise AtError(msg) from exc

            orbit = orbit + np.hstack((dd0, 1.0, 0.0)) * dp * dp
            dorbit = np.hstack((d0 + dd0 * dp, 1.0, 0.0))

            if (rmat is not None) and (drmat is not None):
                rmat = rmat + drmat * dp
            else:
                alphas = alphas + dalpha * dp
                betas = betas + dbeta * dp

        if rmat is not None:
            # For some reason, "emittances" must be different...
            sigm = rmat[0, ...] + 10.0 * rmat[1, ...]
            if rmat.shape[0] >= 3:
                sigm = sigm + 0.1 * rmat[2, ...]
                dorbit = rmat[2, :, 4] / rmat[2, 4, 4, np.newaxis]
        else:
            slices = [slice(2 * i, 2 * (i + 1)) for i in range(2)]
            ab = np.stack((alphas, betas), axis=1)
            sigm = np.zeros((4, 4))
            for slc, (alpha, beta) in zip(slices, ab):
                gamma = (1.0 + alpha * alpha) / beta
                sigm[slc, slc] = np.array([[beta, -alpha], [-alpha, gamma]])

        return orbit, sigm, dorbit

    def chrom_w(ringup, ringdn, orbitup, orbitdn, refpts=None, **kwargs):
        """Compute the chromaticity and W-functions"""

        # noinspection PyShadowingNames
        def off_momentum(rng, orb0, dp=None, **kwargs):
            if twiss_in is None:
                mt, ms = get_matrix(rng, refpts=refpts, orbit=orb0, **kwargs)
                mxx = mt
                dpup = orb0[4] + 0.5 * dp_step
                dpdn = orb0[4] - 0.5 * dp_step
                o0up = None
                o0dn = None
            else:
                orbit, sigma, dorbit = build_sigma(orb0, dp=dp)
                mt, ms = get_matrix(ring, refpts=refpts, orbit=orbit, **kwargs)
                mxx = sigma @ jmat(sigma.shape[0] // 2)
                o0up = orbit + dorbit * 0.5 * dp_step
                o0dn = orbit - dorbit * 0.5 * dp_step
                dpup = None
                dpdn = None
            vps, _, el0, els, wtype = analyze(mxx, ms)
            tunes = _tunes(rng, orbit=orb0)
            o0up, oup = get_orbit(
                ring, refpts=refpts, guess=orb0, dp=dpup, orbit=o0up, **kwargs
            )
            o0dn, odn = get_orbit(
                ring, refpts=refpts, guess=orb0, dp=dpdn, orbit=o0dn, **kwargs
            )
            d0 = (o0up - o0dn)[:4] / dp_step
            ds = np.array([(up - dn)[:4] / dp_step for up, dn in zip(oup, odn)])
            return tunes, el0, els, d0, ds, wtype

        def wget(ddp, elup, eldn, has_r):
            """Compute the chromatic amplitude function"""
            (*data_up,) = elup  # Extract alpha and beta
            (*data_dn,) = eldn
            alpha_up, beta_up, mu_up = data_up[:3]
            alpha_dn, beta_dn, mu_dn = data_dn[:3]
            db = np.array(beta_up - beta_dn) / ddp
            mb = (beta_up + beta_dn) / 2
            da = np.array(alpha_up - alpha_dn) / ddp
            ma = (alpha_up + alpha_dn) / 2
            wa = da - ma / mb * db
            wb = db / mb
            ww = np.sqrt(wa**2 + wb**2)
            wp = np.arctan2(wa, wb)
            dmu = np.array(mu_up - mu_dn) / ddp
            data_out = (ww, wp, da, db, dmu)
            if has_r:
                r_up = data_up[3]
                r_dn = data_dn[3]
                data_out += (np.array(r_up - r_dn) / ddp,)
            return data_out

        deltap = orbitup[4] - orbitdn[4]
        (*data_up,) = off_momentum(ringup, orbitup, dp=0.5 * deltap, **kwargs)
        (*data_dn,) = off_momentum(ringdn, orbitdn, dp=-0.5 * deltap, **kwargs)
        tunesup, el0up, elsup, d0up, dsup, wtype = data_up
        tunesdn, el0dn, elsdn, d0dn, dsdn, _ = data_dn
        has_r = len(wtype) == 7
        # in 6D, dp comes out of find_orbit6
        chrom = (tunesup - tunesdn) / deltap
        dd0 = np.array(d0up - d0dn) / deltap
        dds = np.array(dsup - dsdn) / deltap
        data0 = wget(deltap, el0up, el0dn, has_r)
        datas = wget(deltap, elsup, elsdn, has_r)
        data0 = data0 + (dd0,)
        datas = datas + (dds,)
        return chrom, data0, datas

    def unwrap(mu):
        """Remove the phase jumps"""
        dmu = np.diff(np.concatenate((np.zeros((1, dms)), mu)), axis=0)
        jumps = dmu < -1.0e-3
        mu += np.cumsum(jumps, axis=0) * 2.0 * np.pi

    dp_step = kwargs.get("DPStep", DConstant.DPStep)
    addtype = kwargs.pop("addtype", [])

    if ring.is_6d:
        get_matrix = find_m66
        get_orbit = find_orbit6
    else:
        get_matrix = find_m44
        get_orbit = find_orbit4

    o0up = None
    o0dn = None
    if twiss_in is None:  # Ring
        if orbit is None:
            orbit, _ = get_orbit(
                ring, dp=dp, dct=dct, df=df, keep_lattice=keep_lattice, **kwargs
            )
            keep_lattice = True
        # Get 1-turn transfer matrix
        mt, ms = get_matrix(ring, refpts=refpts, orbit=orbit, **kwargs)
        mxx = mt
    else:  # Transfer line
        orbit, sigma, dorbit = build_sigma(orbit)
        # Get 1-turn transfer matrix
        mt, ms = get_matrix(ring, refpts=refpts, orbit=orbit, **kwargs)
        mxx = sigma @ jmat(sigma.shape[0] // 2)
        o0up = orbit + dorbit * 0.5 * dp_step
        o0dn = orbit - dorbit * 0.5 * dp_step

    # Perform analysis
    vps, dtype, el0, els, wtype = analyze(mxx, ms)
    tunes = np.mod(np.angle(vps) / 2.0 / pi, 1.0)

    if (get_chrom or get_w) and mt.shape == (6, 6):
        dff = dp_step * ring.disable_6d(copy=True).slip_factor
        cavities = get_bool_index(ring, RFCavity)
        freqs = get_value_refpts(ring, cavities, "Frequency")
        rgup = set_value_refpts(
            ring, cavities, "Frequency", freqs * (1.0 + 0.5 * dff), copy=True
        )
        rgdn = set_value_refpts(
            ring, cavities, "Frequency", freqs * (1.0 - 0.5 * dff), copy=True
        )
        if o0up is None:
            o0up, _ = get_orbit(rgup, guess=orbit, orbit=o0up, **kwargs)
            o0dn, _ = get_orbit(rgdn, guess=orbit, orbit=o0dn, **kwargs)
    else:
        rgup = ring
        rgdn = ring

    # Propagate the closed orbit
    orb0, orbs = get_orbit(ring, refpts=refpts, orbit=orbit, keep_lattice=keep_lattice)
    spos = ring.get_s_pos(refpts)

    nrefs = orbs.shape[0]
    dms = vps.size
    if dms >= 3:  # 6D processing
        dtype = dtype + [
            ("closed_orbit", np.float64, (6,)),
            ("M", np.float64, (2 * dms, 2 * dms)),
            ("s_pos", np.float64),
        ]
        data0 = (orb0, np.identity(2 * dms), 0.0)
        datas = (orbs, ms, spos)
        length = ring.get_s_pos(len(ring))[0]
        damping_rates = -np.log(np.absolute(vps))
        damping_times = length / clight / damping_rates
    else:  # 4D processing
        kwargs["keep_lattice"] = True
        dpup = orb0[4] + 0.5 * dp_step
        dpdn = orb0[4] - 0.5 * dp_step
        o0up, oup = get_orbit(
            ring, refpts=refpts, guess=orb0, dp=dpup, orbit=o0up, **kwargs
        )
        o0dn, odn = get_orbit(
            ring, refpts=refpts, guess=orb0, dp=dpdn, orbit=o0dn, **kwargs
        )
        d0 = (o0up - o0dn)[:4] / dp_step
        ds = np.array([(up - dn)[:4] / dp_step for up, dn in zip(oup, odn)])
        dtype = dtype + [
            ("dispersion", np.float64, (4,)),
            ("closed_orbit", np.float64, (6,)),
            (mname, np.float64, (2 * dms, 2 * dms)),
            ("s_pos", np.float64),
        ]
        data0 = (d0, orb0, mt, get_s_pos(ring, len(ring))[0])
        datas = (ds, orbs, ms, spos)
        damping_times = np.nan

    if get_w:
        dtype = dtype + wtype
        chrom, ddata0, ddatas = chrom_w(rgup, rgdn, o0up, o0dn, refpts, **kwargs)
        data0 = data0 + ddata0
        datas = datas + ddatas
    elif get_chrom:
        tunesup = _tunes(rgup, orbit=o0up)
        tunesdn = _tunes(rgdn, orbit=o0dn)
        deltap = o0up[4] - o0dn[4]
        chrom = (tunesup - tunesdn) / deltap
    else:
        chrom = np.nan

    beamdata = np.array(
        (tunes, chrom, damping_times),
        dtype=[
            ("tune", np.float64, (dms,)),
            ("chromaticity", np.float64, (dms,)),
            ("damping_time", np.float64, (dms,)),
        ],
    ).view(np.recarray)

    dtype = dtype + addtype
    elemdata0 = np.array(el0 + data0 + add0, dtype=dtype).view(np.recarray)
    elemdata = np.recarray((nrefs,), dtype=dtype)
    if nrefs > 0:
        for name, value in zip(np.dtype(dtype).names, els + datas + adds):
            elemdata[name] = value
        unwrap(elemdata.mu)
    return elemdata0, beamdata, elemdata


@check_6d(False)
def linopt2(ring: Lattice, *args, **kwargs):
    r"""Linear analysis of an uncoupled lattice

    :py:func:`linopt2` computes the linear optics parameters on a single cell
    (*periodicity* is not taken into account).

    Parameters:
        ring:   Lattice description.

    Keyword Args:
        refpts (Refpts):        Elements at which data is returned.
          It can be:

          1. an integer in the range [-len(ring), len(ring)-1]
             selecting the element according to python indexing rules.
             As a special case, len(ring) is allowed and refers to the end
             of the last element,
          2. an ordered list of such integers without duplicates,
          3. a numpy array of booleans of maximum length len(ring)+1,
             where selected elements are :py:obj:`True`.
        dp (float):             Momentum deviation. Defaults to :py:obj:`None`
        dct (float):            Path lengthening. Defaults to :py:obj:`None`
        df (float):             Deviation of RF frequency. Defaults to
          :py:obj:`None`
        orbit (Orbit):          Avoids looking for the closed orbit if it is
          already known ((6,) array)
        get_chrom (bool):       Compute chromaticities. Needs computing
          the tune at 2 different momentum deviations around the central one.
        get_w (bool):           Computes chromatic amplitude functions
          (W, WP) [4]_, and derivatives of the dispersion and twiss parameters
          versus dp. Needs to compute the optics at 2 different momentum
          deviations around the central one.
        keep_lattice (bool):    Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):         Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        twiss_in:               Initial conditions for transfer line optics.
          Record array as output by :py:func:`.linopt6`, or dictionary. Keys:

          R or alpha, beta
            mandatory (2,) arrays
          closed_orbit
            Optional (6,) array, default 0
          dispersion
            Optional (6,) array, default 0

          If present, the attribute **R** will be used, otherwise the
          attributes **alpha** and **beta** will be used. All other attributes
          are ignored.

    Returns:
        elemdata0:      Linear optics data at the entrance of the ring
        ringdata:       Lattice properties
        elemdata:       Linear optics at the points referred to by *refpts*,
          if refpts is :py:obj:`None` an empty lindata structure is returned.

    .. _linopt2_elemdata:

    **elemdata** is a record array with fields:

    ================    ===================================================
    **s_pos**           longitudinal position [m]
    **M**               (4, 4) transfer matrix M from the beginning of ring
                        to the entrance of the element [2]_
    **closed_orbit**    (6,) closed orbit vector
    **dispersion**      (4,) dispersion vector
    **beta**            :math:`\left[ \beta_x,\beta_y \right]` vector
    **alpha**           :math:`\left[ \alpha_x,\alpha_y \right]` vector
    **mu**              :math:`\left[ \mu_x,\mu_y \right]`, betatron phase
                        (modulo :math:`2\pi`)
    **W**               :math:`\left[ W_x,W_y \right]` only if *get_w*
                        is :py:obj:`True`: chromatic amplitude function
    **Wp**              :math:`\left[ Wp_x,Wp_y \right]` only if *get_w*
                        is :py:obj:`True`: chromatic phase function
    **dalpha**          (2,) alpha derivative vector
                        (:math:`\Delta \alpha/ \delta_p`)
    **dbeta**           (2,) beta derivative vector
                        (:math:`\Delta \beta/ \delta_p`)
    **dmu**             (2,) mu derivative vector
                        (:math:`\Delta \mu/ \delta_p`)
    **ddispersion**     (4,) dispersion derivative vector
                        (:math:`\Delta D/ \delta_p`)
    ================    ===================================================

    All values given at the entrance of each element specified in refpts.
    Field values can be obtained with either *lindata['idx']* or *lindata.idx*

    **ringdata** is a record array with fields:

    =================   ======
    **tune**            Fractional tunes
    **chromaticity**    Chromaticities, only computed if *get_chrom* is
                        :py:obj:`True`
    =================   ======

    References:
        **[1]** D.Edwards,L.Teng IEEE Trans.Nucl.Sci. NS-20, No.3, p.885-888
        , 1973

        .. [2] E.Courant, H.Snyder

        **[3]** D.Sagan, D.Rubin Phys.Rev.Spec.Top.-Accelerators and beams,
        vol.2 (1999)

        .. [4] Brian W. Montague Report LEP Note 165, CERN, 1979
    """
    return _linopt(ring, _analyze2, *args, **kwargs)


@check_6d(False)
def linopt4(ring: Lattice, *args, **kwargs):
    r"""Linear analysis of a H/V coupled lattice

    4D-analysis of coupled motion following Sagan/Rubin [7]_

    :py:func:`linopt4` computes the linear optics parameters on a single cell
    (*periodicity* is not taken into account).

    Parameters:
        ring:   Lattice description.

    Keyword Args:
        refpts (Refpts):        Elements at which data is returned.
          It can be:

          1. an integer in the range [-len(ring), len(ring)-1]
             selecting the element according to python indexing rules.
             As a special case, len(ring) is allowed and refers to the end
             of the last element,
          2. an ordered list of such integers without duplicates,
          3. a numpy array of booleans of maximum length len(ring)+1,
             where selected elements are :py:obj:`True`.
        dp (float):             Momentum deviation. Defaults to :py:obj:`None`
        dct (float):            Path lengthening. Defaults to :py:obj:`None`
        df (float):             Deviation of RF frequency. Defaults to
          :py:obj:`None`
        orbit (Orbit):          Avoids looking for the closed orbit if it is
          already known ((6,) array)
        get_chrom (bool):       Compute chromaticities. Needs computing
          the tune at 2 different momentum deviations around the central one.
        get_w (bool):           Computes chromatic amplitude functions
          (W) [8]_. Needs to compute the optics at 2 different momentum
          deviations around the central one.
        keep_lattice (bool):    Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):         Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        twiss_in:               Initial conditions for transfer line optics.
          Record array as output by :py:func:`.linopt6`, or dictionary. Keys:

          R or alpha, beta
            mandatory (2,) arrays
          closed_orbit
            Optional (6,) array, default 0
          dispersion
            Optional (6,) array, default 0

          If present, the attribute **R** will be used, otherwise the
          attributes **alpha** and **beta** will be used. All other attributes
          are ignored.

    Returns:
        elemdata0:      Linear optics data at the entrance of the ring
        ringdata:       Lattice properties
        elemdata:       Linear optics at the points refered to by *refpts*,
          if refpts is :py:obj:`None` an empty lindata structure is returned.

    .. _linopt4_elemdata:

    **elemdata** is a record array with fields:

    ================    ===================================================
    **s_pos**           longitudinal position [m]
    **M**               (4, 4) transfer matrix M from the beginning of ring
                        to the entrance of the element [6]_
    **closed_orbit**    (6,) closed orbit vector
    **dispersion**      (4,) dispersion vector
    **beta**            :math:`\left[ \beta_x,\beta_y \right]` vector
    **alpha**           :math:`\left[ \alpha_x,\alpha_y \right]` vector
    **mu**              :math:`\left[ \mu_x,\mu_y \right]`, betatron phase
                        (modulo :math:`2\pi`)
    **gamma**           gamma parameter of the transformation to
                        eigenmodes [7]_
    **W**               :math:`\left[ W_x,W_y \right]` only if *get_w*
                        is :py:obj:`True`: chromatic amplitude function
    **Wp**              :math:`\left[ Wp_x,Wp_y \right]` only if *get_w*
                        is :py:obj:`True`: chromatic phase function
    **dalpha**          (2,) alpha derivative vector
                        (:math:`\Delta \alpha/ \delta_p`)
    **dbeta**           (2,) beta derivative vector
                        (:math:`\Delta \beta/ \delta_p`)
    **dmu**             (2,) mu derivative vector
                        (:math:`\Delta \mu/ \delta_p`)
    **ddispersion**     (4,) dispersion derivative vector
                        (:math:`\Delta D/ \delta_p`)
    ================    ===================================================

    All values given at the entrance of each element specified in refpts.
    Field values can be obtained with either
    *lindata['idx']* or *lindata.idx*

    **ringdata** is a record array with fields:

    =================   ======
    **tune**            Fractional tunes
    **chromaticity**    Chromaticities, only computed if *get_chrom* is
                        :py:obj:`True`
    =================   ======

    References:
        **[5]** D.Edwards,L.Teng IEEE Trans.Nucl.Sci. NS-20, No.3, p.885-888
        , 1973

        .. [6] E.Courant, H.Snyder

        .. [7] D.Sagan, D.Rubin Phys.Rev.Spec.Top.-Accelerators and beams,
           vol.2 (1999)

        .. [8] Brian W. Montague Report LEP Note 165, CERN, 1979
    """
    return _linopt(ring, _analyze4, *args, **kwargs)


@frequency_control
def linopt6(ring: Lattice, *args, **kwargs):
    r"""Linear analysis of a fully coupled lattice using normal modes

    For circular machines, :py:func:`linopt6` analyses

    * the 4x4 1-turn transfer matrix if *ring* is 4D, or
    * the 6x6 1-turn transfer matrix if  *ring* is 6D.

    For a transfer line, The "twiss_in" intput must contain either:

    *  a field **R**, as provided by ATLINOPT6, or
    * the fields **beta** and **alpha**, as provided by linopt and linopt6

    :py:func:`linopt6` computes the linear optics parameters on a single cell
    (*periodicity* is not taken into account).

    Parameters:
        ring:   Lattice description.

    Keyword Args:
        refpts (Refpts):        Elements at which data is returned.
          It can be:

          1. an integer in the range [-len(ring), len(ring)-1]
             selecting the element according to python indexing rules.
             As a special case, len(ring) is allowed and refers to the end
             of the last element,
          2. an ordered list of such integers without duplicates,
          3. a numpy array of booleans of maximum length len(ring)+1,
             where selected elements are :py:obj:`True`.
        dp (float):             Momentum deviation. Defaults to :py:obj:`None`
        dct (float):            Path lengthening. Defaults to :py:obj:`None`
        df (float):             Deviation of RF frequency. Defaults to
          :py:obj:`None`
        orbit (Orbit):          Avoids looking for the closed orbit if it is
          already known ((6,) array)
        get_chrom (bool):       Compute chromaticities. Needs computing
          the tune at 2 different momentum deviations around the central one.
        get_w (bool):           Computes chromatic amplitude functions
          (W) [11]_. Needs to compute the optics at 2 different momentum
          deviations around the central one.
        keep_lattice (bool):    Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):         Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        twiss_in:               Initial conditions for transfer line optics.
          Record array as output by :py:func:`.linopt6`, or dictionary. Keys:

          R or alpha, beta
            mandatory (2,) arrays
          closed_orbit
            Optional (6,) array, default 0
          dispersion
            Optional (6,) array, default 0

          If present, the attribute **R** will be used, otherwise the
          attributes **alpha** and **beta** will be used. All other attributes
          are ignored.
        cavpts (Refpts):        Cavity location for off-momentum tuning

    Returns:
        elemdata0:      Linear optics data at the entrance of the ring
        ringdata:       Lattice properties
        elemdata:       Linear optics at the points refered to by *refpts*,
          if refpts is :py:obj:`None` an empty lindata structure is returned.

    .. _linopt6_elemdata:

    **elemdata** is a record array with fields:

    ================    ===================================================
    **s_pos**           longitudinal position [m]
    **M**               (6, 6) transfer matrix M from the beginning of ring
                        to the entrance of the element
    **closed_orbit**    (6,) closed orbit vector
    **dispersion**      (4,) dispersion vector
    **A**               (6,6) A-matrix
    **R**               (3, 6, 6) R-matrices
    **beta**            :math:`\left[ \beta_x,\beta_y \right]` vector
    **alpha**           :math:`\left[ \alpha_x,\alpha_y \right]` vector
    **mu**              :math:`\left[ \mu_x,\mu_y \right]`, betatron phase
                        (modulo :math:`2\pi`)
    **W**               :math:`\left[ W_x,W_y \right]` only if *get_w*
                        is :py:obj:`True`: chromatic amplitude function
    **Wp**              :math:`\left[ Wp_x,Wp_y \right]` only if *get_w*
                        is :py:obj:`True`: chromatic phase function
    **dalpha**          (2,) alpha derivative vector
                        (:math:`\Delta \alpha/ \delta_p`)
    **dbeta**           (2,) beta derivative vector
                        (:math:`\Delta \beta/ \delta_p`)
    **dmu**             (2,) mu derivative vector
                        (:math:`\Delta \mu/ \delta_p`)
    **ddispersion**     (4,) dispersion derivative vector
                        (:math:`\Delta D/ \delta_p`)
    **dR**              (3, 6, 6) R derivative vector
                        (:math:`\Delta R/ \delta_p`)
    ================    ===================================================

    All values given at the entrance of each element specified in refpts.
    Field values can be obtained with either
    *lindata['idx']* or *lindata.idx*

    **ringdata** is a record array with fields:

    =================   ======
    **tune**            Fractional tunes
    **chromaticity**    Chromaticities, only computed if *get_chrom* is
                        :py:obj:`True`
    **damping_time**    Damping times [s] (only if radiation is ON)
    =================   ======

    References:
        **[9]** Etienne Forest, Phys. Rev. E 58, 2481 – Published 1 August 1998

        **[10]** Andrzej Wolski, Phys. Rev. ST Accel. Beams 9, 024001 –
        Published 3 February 2006

        .. [11] Brian W. Montague Report LEP Note 165, CERN, 1979
    """
    return _linopt(ring, _analyze6, *args, **kwargs)


def linopt_auto(ring: Lattice, *args, **kwargs):
    """
    This is a convenience function to automatically switch to the faster
    :py:func:`linopt2` in case the *coupled* keyword argument is
    :py:obj:`False` **and** ring.is_6d is :py:obj:`False`.
    Otherwise, the default :py:func:`linopt6` is used

    Parameters: Same as :py:func:`.linopt2` or :py:func:`.linopt6`

    Keyword Args;
        coupled (bool):     If set to :py:obj:`False`, H/V coupling
          will be ignored to simplify the calculation (needs ring.is_6d
          :py:obj:`False`)


    Returns:
        elemdata0:      Linear optics data at the entrance of the ring
        ringdata:       Lattice properties
        elemdata:       Linear optics at the points referred to by *refpts*,
          if refpts is :py:obj:`None` an empty lindata structure is returned.

    Warning:
        The output varies depending whether :py:func:`.linopt2` or
        :py:func:`.linopt6` is called. To be used with care!
    """
    if not (kwargs.pop("coupled", True) or ring.is_6d):
        return linopt2(ring, *args, **kwargs)
    else:
        return linopt6(ring, *args, **kwargs)


def get_optics(
    ring: Lattice,
    refpts: Refpts = None,
    dp: float = None,
    method: Callable = linopt6,
    **kwargs,
):
    """Linear analysis of a fully coupled lattice

    :py:func:`get_optics` computes the linear optics parameters on a single cell
    (*periodicity* is not taken into account).

    Parameters:
        ring:                   Lattice description.
        refpts:                 Elements at which data is returned.
          It can be:

          1. an integer in the range [-len(ring), len(ring)-1]
             selecting the element according to python indexing rules.
             As a special case, len(ring) is allowed and refers to the end
             of the last element,
          2. an ordered list of such integers without duplicates,
          3. a numpy array of booleans of maximum length len(ring)+1,
             where selected elements are :py:obj:`True`.
        dp:                     Momentum deviation.
        method (Callable):      Method for linear optics:

          :py:obj:`~.linear.linopt2`: no longitudinal motion, no H/V coupling,

          :py:obj:`~.linear.linopt4`: no longitudinal motion, Sagan/Rubin
          4D-analysis of coupled motion,

          :py:obj:`~.linear.linopt6` (default): with or without longitudinal
          motion, normal mode analysis

    Keyword Args:
        dct (float):            Path lengthening. Defaults to :py:obj:`None`
        df (float):             Deviation of RF frequency. Defaults to
          :py:obj:`None`
        orbit (Orbit):          Avoids looking for the closed orbit if it is
          already known ((6,) array)
        get_chrom (bool):       Compute chromaticities. Needs computing
          the tune at 2 different momentum deviations around the central one.
        get_w (bool):           Computes chromatic amplitude functions
          (W) [4]_. Needs to compute the optics at 2 different momentum
          deviations around the central one.
        keep_lattice (bool):    Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):         Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        twiss_in:               Initial conditions for transfer line optics.
          Record array as output by :py:func:`.linopt6`, or dictionary. Keys:

          R or alpha, beta
            mandatory (2,) arrays
          closed_orbit
            Optional (6,) array, default 0
          dispersion
            Optional (6,) array, default 0

          If present, the attribute **R** will be used, otherwise the
          attributes **alpha** and **beta** will be used. All other attributes
          are ignored.

    Returns:
        elemdata0:      Linear optics data at the entrance of the ring
        ringdata:       Lattice properties
        elemdata:       Linear optics at the points refered to by *refpts*,
          if refpts is :py:obj:`None` an empty lindata structure is returned.


    **elemdata** is a record array, Its fields depend on the selected method. See
    :ref:`linopt6 <linopt6_elemdata>`, :ref:`linopt4 <linopt4_elemdata>`,
    :ref:`linopt2 <linopt2_elemdata>`

    **ringdata** is a record array with fields:

    =================   ======
    **tune**            Fractional tunes
    **chromaticity**    Chromaticities, only computed if *get_chrom* is
                        :py:obj:`True`
    **damping_time**    Damping times [s] (only if radiation is ON)
    =================   ======


    Warning:
        The format of output record arrays depends on the selected method.
        See :py:func:`linopt2`, :py:func:`linopt4`, :py:func:`linopt6`.
    """
    return method(ring, refpts=refpts, dp=dp, **kwargs)


# noinspection PyPep8Naming
@check_6d(False)
def linopt(
    ring: Lattice,
    dp: float = 0.0,
    refpts: Refpts = None,
    get_chrom: bool = False,
    **kwargs,
):
    """Linear analysis of a H/V coupled lattice (deprecated)

    Parameters:
        ring:           lattice description.
        dp:             momentum deviation.
        refpts:         elements at which data is returned. It can be:
                        1) an integer in the range [-len(ring), len(ring)-1]
                           selecting the element according to python indexing
                           rules. As a special case, len(ring) is allowed and
                           refers to the end of the last element,
                        2) an ordered list of such integers without duplicates,
                        3) a numpy array of booleans of maximum length
                           len(ring)+1, where selected elements are
                           :py:obj:`True`.
    Keyword Args:
        orbit           avoids looking for the closed orbit if is already known
                        ((6,) array)
        get_chrom=False compute chromaticities. Needs computing the tune at
                        2 different momentum deviations around the central one.
        get_w=False     computes chromatic amplitude functions (W) [4].
                        Needs to compute the optics at 2 different momentum
                        deviations around the central one.
        keep_lattice    Assume no lattice change since the previous tracking.
                        Defaults to :py:obj:`False`
        XYStep=1.0e-8   transverse step for numerical computation
        DPStep=1.0E-6   momentum deviation used for computation of
                        chromaticities and dispersion
        coupled=True    if :py:obj:`False`, simplify the calculations by
                        assuming no H/V coupling
        twiss_in=None   Initial conditions for transfer line optics. Record
                        array as output by linopt, or dictionary. Keys:
                        'alpha' and 'beta'  (mandatory)
                        'closed_orbit',     (default 0)
                        'dispersion'        (default 0)
                        All other attributes are ignored.
    OUTPUT
        lindata0        linear optics data at the entrance of the ring
        tune            [tune_A, tune_B], linear tunes for the two normal modes
                        of linear motion [1]
        chrom           [ksi_A , ksi_B], chromaticities ksi = d(nu)/(dP/P).
                        Only computed if 'get_chrom' is :py:obj:`True`
        lindata         linear optics at the points referred to by refpts, if
                        refpts is None an empty lindata structure is returned.

        lindata is a record array with fields:
        idx             element index in the ring
        s_pos           longitudinal position [m]
        m44             (4, 4) transfer matrix M from the beginning of ring
                        to the entrance of the element [2]
        closed_orbit    (6,) closed orbit vector
        dispersion      (4,) dispersion vector
        beta            [betax, betay] vector
        alpha           [alphax, alphay] vector
        mu              [mux, muy], betatron phase (modulo 2*pi)
        W               (2,) chromatic amplitude function (only if get_w==True)
        All values given at the entrance of each element specified in refpts.
        In case coupled == :py:obj:`True` additional outputs are available:
        gamma           gamma parameter of the transformation to eigenmodes
        A               (2, 2) matrix A in [3]
        B               (2, 2) matrix B in [3]
        C               (2, 2) matrix C in [3]
        Field values can be obtained with either
        lindata['idx']    or
        lindata.idx
    REFERENCES
        [1] D.Edwards,L.Teng IEEE Trans.Nucl.Sci. NS-20, No.3, p.885-888, 1973
        [2] E.Courant, H.Snyder
        [3] D.Sagan, D.Rubin Phys.Rev.Spec.Top.-Accelerators and beams,
            vol.2 (1999)
        [4] Brian W. Montague Report LEP Note 165, CERN, 1979

    :meta private:
    """
    analyze = _analyze4 if kwargs.pop("coupled", True) else _analyze2
    eld0, bd, eld = _linopt(
        ring,
        analyze,
        refpts,
        dp=dp,
        get_chrom=get_chrom,
        add0=(0,),
        adds=(get_uint32_index(ring, refpts),),
        addtype=[("idx", np.uint32)],
        mname="m44",
        **kwargs,
    )
    # noinspection PyUnresolvedReferences
    return eld0, bd.tune, bd.chromaticity, eld


# noinspection PyPep8Naming
@check_6d(False)
def avlinopt(ring: Lattice, dp: float = 0.0, refpts: Refpts = None, **kwargs):
    r"""Linear analysis of a lattice with average values

    :py:func:`avlinopt` returns average beta, mu, dispersion over the lattice
    elements.

    :py:func:`avlinopt` returns average beta, mu, dispersion over the lattice
    elements. :py:func:`avlinopt` acts on a single cell (*periodicity* is not taken
    into account).

    Parameters:
        ring:       Lattice description.
        dp:         Momentum deviation.
        refpts:     Elements at which data is returned.
          It can be:

          1. an integer in the range [-len(ring), len(ring)-1]
             selecting the element according to python indexing rules.
             As a special case, len(ring) is allowed and refers to the end
             of the last element,
          2. an ordered list of such integers without duplicates,
          3. a numpy array of booleans of maximum length len(ring)+1,
             where selected elements are :py:obj:`True`.

    Keyword Args:
        dct (float):            Path lengthening. Defaults to :py:obj:`None`
        df (float):             Deviation of RF frequency. Defaults to
          :py:obj:`None`
        orbit (Orbit):          Avoids looking for the closed orbit if it is
          already known ((6,) array)
        get_chrom (bool):       Compute chromaticities. Needs computing
          the tune at 2 different momentum deviations around the central one.
        get_w (bool):           Computes chromatic amplitude functions
          (W) [4]_. Needs to compute the optics at 2 different momentum
          deviations around the central one.
        keep_lattice (bool):    Assume no lattice change since the
          previous tracking. Defaults to :py:obj:`False`
        XYStep (float):         Step size.
          Default: :py:data:`DConstant.XYStep <.DConstant>`
        DPStep (float):         Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`
        twiss_in:               Initial conditions for transfer line optics.
          Record array as output by :py:func:`.linopt6`, or dictionary. Keys:

          R or alpha, beta
            mandatory (2,) arrays
          closed_orbit
            Optional (6,) array, default 0
          dispersion
            Optional (6,) array, default 0

          If present, the attribute **R** will be used, otherwise the
          attributes **alpha** and **beta** will be used. All other attributes
          are ignored.

    Returns:
        elemdata:   Linear optics at the points referred to by *refpts*,
          if refpts is :py:obj:`None` an empty lindata structure is returned.
        avebeta:    Average beta functions
          [:math:`\hat{\beta_x},\hat{\beta_y}`] at *refpts*
        avemu:      Average phase advances
          [:math:`\hat{\mu_x},\hat{\mu_y}`] at *refpts*
        avedisp:    Average dispersion [:math:`\hat{\eta_x}, \hat{\eta'_x},
          \hat{\eta_y}, \hat{\eta'_y}`] at *refpts*
        avespos:    Average s position at *refpts*
        tune:       [:math:`\nu_1,\nu_2`], linear tunes for the two normal
          modes of linear motion [1]
        chrom:      [:math:`\xi_1,\xi_2`], chromaticities

    See also:
        :py:func:`linopt4`, :py:func:`get_optics`
    """

    def get_strength(elem):
        try:
            k = elem.PolynomB[1]
        except (AttributeError, IndexError):
            k = 0.0
        return k

    def get_sext_strength(elem):
        try:
            k = elem.PolynomB[2]
        except (AttributeError, IndexError):
            k = 0.0
        return k

    def get_roll(elem):
        try:
            k = elem.R2[0][0]
        except (AttributeError, IndexError):
            k = 1.0
        return k

    def get_dx(elem):
        try:
            k = (elem.T2[0] - elem.T1[0]) / 2.0
        except (AttributeError, IndexError):
            k = 0.0
        return k

    def get_bendingangle(elem):
        try:
            k = elem.BendingAngle
        except AttributeError:
            k = 0.0
        return k

    def get_e1(elem):
        try:
            k = elem.EntranceAngle
        except AttributeError:
            k = 0.0
        return k

    def get_fint(elem):
        try:
            k = elem.FringeInt1
        except AttributeError:
            k = 0.0
        return k

    def get_gap(elem):
        try:
            k = elem.FullGap
        except AttributeError:
            k = 0.0
        return k

    def foctrigo(k2, lg):
        sqkl = np.sqrt(k2) * lg
        return np.sin(sqkl) / sqkl, 1.0 - np.cos(sqkl)

    def defoctrigo(k2, lg):
        sqkl = np.sqrt(-k2) * lg
        return np.sinh(sqkl) / sqkl, 1.0 - np.cosh(sqkl)

    def betadrift(beta0, alpha0, lg):
        gamma0 = (alpha0 * alpha0 + 1.0) / beta0
        return beta0 - alpha0 * lg + gamma0 * lg * lg / 3

    def betafoc(func, beta0, alpha0, k2, lg):
        gamma0 = (alpha0 * alpha0 + 1.0) / beta0
        sini, cosi = func(k2, 2.0 * lg)
        return (
            (beta0 + gamma0 / k2)
            + (beta0 - gamma0 / k2) * sini
            - cosi * alpha0 / k2 / lg
        ) / 2.0

    def betalong(beta0, alpha0, k2, lg):
        r = np.empty_like(beta0)
        kp = k2 >= 1.0e-7
        km = k2 <= -1.0e-7
        k0 = np.logical_not(kp | km)
        r[kp] = betafoc(foctrigo, beta0[kp], alpha0[kp], k2[kp], lg[kp])
        r[km] = betafoc(defoctrigo, beta0[km], alpha0[km], k2[km], lg[km])
        r[k0] = betadrift(beta0[k0], alpha0[k0], lg[k0])
        return r

    def dispfoc(func, disp0, ir, k2, lg):
        sini, cosi = func(k2, lg)
        eta = disp0[:, 0] * sini + disp0[:, 1] * cosi / k2 / lg + ir * (1.0 - sini) / k2
        etap = disp0[:, 1] * sini - disp0[:, 0] * cosi / lg + ir * cosi / k2 / lg
        return np.stack((eta, etap), axis=1)

    def dispdrift(disp0, ir, lg):
        eta = disp0[:, 0] + disp0[:, 1] * lg / 2.0 + lg * lg * ir / 6.0
        etap = disp0[:, 1] + lg * ir / 2.0
        return np.stack((eta, etap), axis=1)

    def displong(disp0, ir, k2, lg):
        r = np.empty_like(disp0)
        kp = k2 >= 1.0e-7
        km = k2 <= -1.0e-7
        k0 = np.logical_not(kp | km)
        r[kp] = dispfoc(foctrigo, disp0[kp], ir[kp], k2[kp], lg[kp])
        r[km] = dispfoc(defoctrigo, disp0[km], ir[km], k2[km], lg[km])
        r[k0] = dispdrift(disp0[k0], ir[k0], lg[k0])
        return r

    # selected list
    boolrefs = get_bool_index(ring, refpts)
    L = np.array([el.Length for el in ring[boolrefs[:-1]]])
    longelem = boolrefs[:-1].copy()
    longelem[longelem] = L != 0.0

    longi_refpts = np.append(longelem, [False])
    longf_refpts = np.roll(longi_refpts, 1)
    all_refs = boolrefs | longf_refpts
    _, bd, d_all = get_optics(ring, refpts=all_refs, dp=dp, get_chrom=True, **kwargs)
    lindata = d_all[boolrefs[all_refs]]  # Optics at entrance of selected elements

    avebeta = lindata.beta.copy()
    avemu = lindata.mu.copy()
    avedisp = lindata.dispersion.copy()
    aves = lindata.s_pos.copy()

    # Long elements
    b_long = longi_refpts[boolrefs]
    ring_long = ring[longelem]
    L = L[(L != 0.0)]
    di = d_all[longi_refpts[all_refs]]  # Optics at entrance of long elements
    df = d_all[longf_refpts[all_refs]]  # Optics at exit of long elements

    dx0 = (di.closed_orbit[:, 0] + df.closed_orbit[:, 0]) / 2
    K = np.array([get_strength(el) for el in ring_long])
    sext_strength = np.array([get_sext_strength(el) for el in ring_long])
    roll = np.array([get_roll(el) for el in ring_long])
    ba = np.array([get_bendingangle(el) for el in ring_long])
    dx = np.array([get_dx(el) for el in ring_long])
    irho = ba / L
    K = K * roll + 2 * sext_strength * (dx0 - dx)
    Kx = K + irho * irho
    Ky = -K
    Kxy = np.stack((Kx, Ky), axis=1)
    Lxy = np.stack((L, L), axis=1)

    # Hard edge model on dipoles
    bend = irho != 0.0
    e1 = np.array([get_e1(el) for el in ring_long[bend]])
    Fint = np.array([get_fint(el) for el in ring_long[bend]])
    gap = np.array([get_gap(el) for el in ring_long[bend]])
    d_csi = irho[bend] * gap * Fint * (1.0 + np.sin(e1) ** 2) / np.cos(e1)
    Cp = np.stack((irho[bend] * np.tan(e1), -irho[bend] * np.tan(e1 - d_csi)), axis=1)
    di.alpha[bend] -= di.beta[bend] * Cp
    di.dispersion[np.ix_(bend, [1, 3])] -= di.dispersion[np.ix_(bend, [0, 2])] * Cp

    avemu[b_long] = 0.5 * (di.mu + df.mu)
    aves[b_long] = 0.5 * (df.s_pos + di.s_pos)
    avebeta[b_long] = betalong(di.beta, di.alpha, Kxy, Lxy)
    avedx = displong(di.dispersion[:, :2], irho, Kx, L)
    avedy = displong(di.dispersion[:, 2:], np.zeros_like(irho), Ky, L)
    avedisp[b_long] = np.concatenate((avedx, avedy), axis=1)

    return lindata, avebeta, avemu, avedisp, aves, bd.tune, bd.chromaticity


@frequency_control
def get_tune(
    ring: Lattice,
    *,
    method: str = "linopt",
    dp: float = None,
    dct: float = None,
    df: float = None,
    orbit: Orbit = None,
    **kwargs,
):
    r"""Computes the tunes using several available methods

    :py:func:`get_tune` may use several methods depending on a *method* keyword.

    Parameters:
        ring:                   Lattice description
        method:                 ``'linopt'`` returns the tunes from the
          :py:func:`.linopt6` function,

          ``'fft'`` tracks a single particle and computes the tunes with fft,

          ``'interp_fft'`` tracks a single particle and computes the tunes with
          interpolated FFT.
        dp (float):             Momentum deviation.
        dct (float):            Path lengthening.
        df (float):             Deviation of RF frequency.
        orbit (Orbit):          Avoids looking for the closed orbit if it is
          already known ((6,) array)

    for the ``'fft'`` and ``'interp_fft'`` methods only:

    Keyword Args:
        nturns (int):           Number of turns. Default: 512
        amplitude (float):      Amplitude of oscillation.
          Default: 1.E-6
        remove_dc (bool):       Remove the mean of oscillation data.
          Default: :py:obj:`True`
        num_harmonics (int):    Number of harmonic components to
          compute (before mask applied, default: 20)
        fmin (float):           Lower tune bound. Default: 0
        fmax (float):           Upper tune bound. Default: 1
        hann (bool):            Turn on Hanning window.
          Default: :py:obj:`False`. Work only for ``'fft'``
        get_integer(bool):   Turn on integer tune (slower)

    Returns:
        tunes (ndarray):        array([:math:`\nu_x,\nu_y`])
    """

    # noinspection PyShadowingNames
    def gen_centroid(ring, ampl, nturns, remove_dc, ld):
        nv = ld.A.shape[0]
        p0 = ld.closed_orbit.copy()
        p0[0] += ampl
        p0[2] += ampl
        if nv >= 6:
            p0[4] += ampl
        p1 = np.squeeze(internal_lpass(ring, p0, nturns, len(ring)))
        if remove_dc:
            p1 -= np.mean(p1, axis=1, keepdims=True)
        p2 = solve(ld.A, p1[:nv, :])
        return np.conjugate(p2.T.view(dtype=complex).T)

    get_integer = kwargs.pop("get_integer", False)
    if get_integer:
        assert method == "linopt", "Integer tune only accessible with method=linopt"
    if method == "linopt":
        if get_integer:
            _, _, c = get_optics(
                ring, refpts=range(len(ring) + 1), dp=dp, dct=dct, df=df, orbit=orbit
            )
            tunes = c.mu[-1] / (2 * np.pi) * ring.periodicity
        else:
            tunes = _tunes(ring, dp=dp, dct=dct, df=df, orbit=orbit)
            tunes, _ = np.modf(tunes * ring.periodicity)
    else:
        nturns = kwargs.pop("nturns", 512)
        ampl = kwargs.pop("ampl", 1.0e-6)
        remove_dc = kwargs.pop("remove_dc", True)
        ld, _, _ = linopt6(ring, dp=dp, dct=dct, df=df, orbit=orbit)
        cents = gen_centroid(ring, ampl, nturns, remove_dc, ld)
        tunes = get_tunes_harmonic(cents, method=method, **kwargs)
        tunes, _ = np.modf(tunes * ring.periodicity)
    return tunes


@frequency_control
def get_chrom(
    ring: Lattice,
    *,
    method: str = "linopt",
    dp: float = None,
    dct: float = None,
    df: float = None,
    cavpts: Refpts = None,
    **kwargs,
):
    r"""Computes the chromaticities using several available methods

    :py:func:`get_tune` may use several methods depending on a *method* keyword.

    Parameters:
        ring:               Lattice description.
        method:             ``'linopt'`` returns the tunes from the
          :py:func:`linopt6` function,

          ``'fft'`` tracks a single particle and computes the tunes with
          :py:func:`~scipy.fftpack.fft`,

          ``'interp_fft'`` tracks a single particle and computes the tunes with
          interpolated FFT.
        dp (float):         Momentum deviation.
        dct (float):        Path lengthening.
        df (float):         Deviation of RF frequency.
        cavpts:     If :py:obj:`None`, look for ring.cavpts, or
          otherwise take all cavities.

    Keyword Args:
        DPStep (float):     Momentum step for differentiation
          Default: :py:data:`DConstant.DPStep <.DConstant>`

    for the ``'fft'`` and ``'interp_fft'`` methods only:

    Keyword Args:
        nturns (int):       Number of turns. Default: 512
        amplitude (float):  Amplitude of oscillation.
          Default: 1.E-6
        remove_dc (bool):   Remove the mean of oscillation data.
          Default: :py:obj:`True`
        num_harmonics (int):Number of harmonic components to
          compute (before mask applied, default: 20)
        fmin (float):       Lower tune bound. Default: 0
        fmax (float):       Upper tune bound. Default: 1
        hann (bool):        Turn on Hanning window.
          Default: :py:obj:`False`, Work only for ``'fft'``

    Returns:
        chromaticities (ndarray):   array([:math:`\xi_x,\xi_y`])
    """

    dp_step = kwargs.pop("DPStep", DConstant.DPStep)
    if method == "fft":
        print("Warning fft method not accurate to get the chromaticity")

    if ring.is_6d:
        dff = dp_step * ring.disable_6d(copy=True).slip_factor
        cavities = get_bool_index(ring, RFCavity)
        freqs = get_value_refpts(ring, cavities, "Frequency")
        rgup = set_value_refpts(
            ring, cavities, "Frequency", freqs * (1.0 + 0.5 * dff), copy=True
        )
        rgdn = set_value_refpts(
            ring, cavities, "Frequency", freqs * (1.0 - 0.5 * dff), copy=True
        )
        o0up, _ = find_orbit6(rgup, **kwargs)
        tune_up = get_tune(rgup, method=method, orbit=o0up, **kwargs)
        o0dn, _ = find_orbit6(rgdn, **kwargs)
        tune_down = get_tune(rgdn, method=method, orbit=o0dn, **kwargs)
        dp_step = o0up[4] - o0dn[4]
    else:
        if dct is not None or df is not None:
            dp = find_orbit4(ring, dct=dct, df=df)[0][4]
        elif dp is None:
            dp = 0.0
        tune_up = get_tune(ring, method=method, dp=dp + 0.5 * dp_step, **kwargs)
        tune_down = get_tune(ring, method=method, dp=dp - 0.5 * dp_step, **kwargs)

    return (tune_up - tune_down) / dp_step


Lattice.linopt = linopt
Lattice.linopt2 = linopt2
Lattice.linopt4 = linopt4
Lattice.linopt6 = linopt6
Lattice.get_optics = get_optics
Lattice.avlinopt = avlinopt
Lattice.get_tune = get_tune
Lattice.get_chrom = get_chrom



"""Resonance Driving Terms"""
from __future__ import annotations

import numpy as np
import multiprocessing
from functools import partial
from ..lattice import Lattice, Refpts, Multipole, All
from enum import Enum
from collections.abc import Container


__all__ = ["get_rdts", "RDTType"]


class RDTType(Enum):
    """Enum class for RDT type"""

    ALL = 0  #: all available RDTs
    FOCUSING = 1  #: Normal quadrupole RDTs
    COUPLING = 2  #: Linear coupling RDTs
    CHROMATIC = 3  #: Chromatic RDTs
    #: Geometric RDTs from sextupoles
    GEOMETRIC1 = 4
    #: Geometric RDTs from octupoles
    #: optionally includes the second order contribution of sextupoles
    GEOMETRIC2 = 5
    #: Amplitude detuning coefficients
    #: optionally includes the second order contribution of sextupoles
    TUNESHIFT = 6


def _get_polynom(elem, attr, index):
    try:
        val = getattr(elem, attr)[index]
        return val
    except (IndexError, AttributeError):
        return 0


def _compute_pf(tune, nperiods):
    """This uses the formula Sum(x^k, k=1->p) = x(x^p-1)/(x-1)"""
    pf = np.ones((9, 9), dtype=complex)
    if nperiods != 1:
        for i in range(9):
            for j in range(9):
                a1 = np.pi * 2 * (tune[0] * (i - 4) + tune[1] * (j - 4))
                a2 = a1 / nperiods
                pf[i][j] = (np.exp(1j * a1) - 1.0) / (np.exp(1j * a2) - 1.0)
    return pf


def _computedrivingterms(
    s,
    betax,
    betay,
    phix,
    phiy,
    etax,
    a2l,
    b2l,
    b3l,
    b4l,
    tune,
    rdttype,
    nperiods,
    refpt,
    second_order,
    periodic_factor,
):
    """
    Original implementation from ELEGANT
    """

    def pf(i, j):
        return periodic_factor[4 + i][4 + j]

    rdts = {}
    rdts2 = {}
    rbetax = np.sqrt(betax)
    rbetay = np.sqrt(betay)
    px = np.exp(1j * phix)
    py = np.exp(1j * phiy)

    rdts["refpts"] = refpt
    mask_a2l = np.absolute(a2l) > 1.0e-6
    mask_b2l = np.absolute(b2l) > 1.0e-6
    mask_b3l = np.absolute(b3l) > 1.0e-6
    mask_b4l = np.absolute(b4l) > 1.0e-6

    if (RDTType.FOCUSING in rdttype) or (RDTType.ALL in rdttype):
        b2lm = b2l[mask_b2l]
        betaxm = betax[mask_b2l]
        betaym = betay[mask_b2l]
        pxm = px[mask_b2l]
        pym = py[mask_b2l]
        rdts["h20000"] = sum((b2lm / 8) * betaxm * pxm * pxm * pf(2, 0))
        rdts["h00200"] = -sum((b2lm / 8) * betaym * pym * pym * pf(0, 2))

    if (RDTType.COUPLING in rdttype) or (RDTType.ALL in rdttype):
        a2lm = a2l[mask_a2l]
        rbetaxm = rbetax[mask_a2l]
        rbetaym = rbetay[mask_a2l]
        pxm = px[mask_a2l]
        pym = py[mask_a2l]
        rdts["h10010"] = sum((a2lm / 4) * rbetaxm * rbetaym * pxm / pym * pf(1, -1))
        rdts["h10100"] = sum((a2lm / 4) * rbetaxm * rbetaym * pxm * pym * pf(1, 1))

    if (RDTType.CHROMATIC in rdttype) or (RDTType.ALL in rdttype):
        mask_b23l = mask_b2l | mask_b3l
        b2lm = b2l[mask_b23l]
        b3lm = b3l[mask_b23l]
        betaxm = betax[mask_b23l]
        rbetaxm = rbetax[mask_b23l]
        betaym = betay[mask_b23l]
        etaxm = etax[mask_b23l]
        pxm = px[mask_b23l]
        pym = py[mask_b23l]
        rdts["h11001"] = sum((b3lm * betaxm * etaxm / 2 - b2lm * betaxm / 4) * nperiods)
        rdts["h00111"] = sum((b2lm * betaym / 4 - b3lm * betaym * etaxm / 2) * nperiods)
        rdts["h20001"] = sum(
            (b3lm * betaxm * etaxm / 2 - b2lm * betaxm / 4) / 2 * pxm * pxm * pf(2, 0)
        )
        rdts["h00201"] = sum(
            (b2lm * betaym / 4 - b3lm * betaym * etaxm / 2) / 2 * pym * pym * pf(0, 2)
        )
        # fmt: off
        rdts["h10002"] = sum(
            (b3lm * rbetaxm * etaxm * etaxm - b2lm * rbetaxm * etaxm)
            / 2 * pxm * pf(1, 0)
        )
        # fmt: on

    if (RDTType.GEOMETRIC1 in rdttype) or (RDTType.ALL in rdttype):
        b3lm = b3l[mask_b3l]
        betaxm = betax[mask_b3l]
        betaym = betay[mask_b3l]
        rbetaxm = rbetax[mask_b3l]
        pxm = px[mask_b3l]
        pym = py[mask_b3l]
        rdts["h21000"] = sum(b3lm * rbetaxm * betaxm / 8 * pxm * pf(1, 0))
        rdts["h30000"] = sum(b3lm * rbetaxm * betaxm / 24 * pxm * pxm * pxm * pf(3, 0))
        rdts["h10110"] = sum(-b3lm * rbetaxm * betaym / 4 * pxm * pf(1, 0))
        rdts["h10020"] = sum(
            -b3lm * rbetaxm * betaym / 8 * pxm * np.conj(pym * pym) * pf(1, -2)
        )
        rdts["h10200"] = sum(-b3lm * rbetaxm * betaym / 8 * pxm * pym * pym * pf(1, 2))

    if (RDTType.TUNESHIFT in rdttype) or (RDTType.ALL in rdttype):
        b4lm = b4l[mask_b4l]
        betaxm = betax[mask_b4l]
        betaym = betay[mask_b4l]
        rdts["dnux_dJx"] = sum(3 * b4lm * betaxm * betaxm / (8 * np.pi) * nperiods)
        rdts["dnux_dJy"] = sum(-3 * b4lm * betaxm * betaym / (4 * np.pi) * nperiods)
        rdts["dnuy_dJy"] = sum(3 * b4lm * betaym * betaym / (8 * np.pi) * nperiods)

    if (RDTType.GEOMETRIC2 in rdttype) or (RDTType.ALL in rdttype):
        b4lm = b4l[mask_b4l]
        betaxm = betax[mask_b4l]
        betaym = betay[mask_b4l]
        pxm = px[mask_b4l]
        pym = py[mask_b4l]
        rdts["h22000"] = sum(3 * b4lm * betaxm * betaxm / 32 * nperiods)
        rdts["h11110"] = sum(-3 * b4lm * betaxm * betaym / 8 * nperiods)
        rdts["h00220"] = sum(3 * b4lm * betaym * betaym / 32 * nperiods)
        rdts["h31000"] = sum(b4lm * betaxm * betaxm / 16 * pxm * pxm * pf(2, 0))
        rdts["h40000"] = sum(
            b4lm * betaxm * betaxm / 64 * pxm * pxm * pxm * pxm * pf(4, 0)
        )
        rdts["h20110"] = sum(-3 * b4lm * betaxm * betaym / 16 * pxm * pxm * pf(2, 0))
        rdts["h11200"] = sum(-3 * b4lm * betaxm * betaym / 16 * pym * pym * pf(0, 2))
        # fmt: off
        rdts["h20020"] = sum(
            -3 * b4lm * betaxm * betaym / 32 * pxm * pxm
            * np.conj(pym * pym) * pf(2, -2)
        )
        # fmt: on
        rdts["h20200"] = sum(
            -3 * b4lm * betaxm * betaym / 32 * pxm * pxm * pym * pym * pf(2, 2)
        )
        rdts["h00310"] = sum(b4lm * betaym * betaym / 16 * pym * pym * pf(0, 2))
        rdts["h00400"] = sum(
            b4lm * betaym * betaym / 64 * pym * pym * pym * pym * pf(0, 4)
        )

    if second_order:
        assert nperiods == 1, "Second order available only for nperiods=1"

        if (RDTType.GEOMETRIC2 in rdttype) or (RDTType.ALL in rdttype):
            nelem = sum(mask_b3l)
            sm = s[mask_b3l]
            b3lm = b3l[mask_b3l]
            betaxm = betax[mask_b3l]
            betaym = betay[mask_b3l]
            rbetaxm = rbetax[mask_b3l]
            rbbetaxm = betaxm * rbetaxm
            phixm = phix[mask_b3l]
            phiym = phiy[mask_b3l]
            pxm = px[mask_b3l]
            pxm2 = pxm * pxm
            pxm3 = pxm2 * pxm
            cpxm = np.conj(pxm)
            cpxm3 = np.conj(pxm3)
            pym = py[mask_b3l]
            pym2 = pym * pym
            cpym2 = np.conj(pym2)
            cpxym2 = np.conj(pxm * pym2)
            rdts2.update(
                {
                    "h22000": 0.0,
                    "h31000": 0.0,
                    "h11110": 0.0,
                    "h11200": 0.0,
                    "h40000": 0.0,
                    "h20020": 0.0,
                    "h20110": 0.0,
                    "h20200": 0.0,
                    "h00220": 0.0,
                    "h00310": 0.0,
                    "h00400": 0.0,
                }
            )
            # fmt: off
            bsign = np.array([np.sign(sm[i] - sm) * 1j * b3lm[i] * b3lm
                              for i in range(nelem)])
            rbbx = np.array([bsign[i] * rbbetaxm[i] * rbbetaxm
                             for i in range(nelem)])
            rbxy = np.array([bsign[i] * rbetaxm[i] * rbetaxm * betaym[i]
                             for i in range(nelem)])
            ppxm = np.array([pxm[i] * pxm for i in range(nelem)])

            rdts2["h22000"] += (1.0 / 64) * np.sum([
                rbbx[i] * (pxm3[i] * cpxm3 + 3 * pxm[i] * cpxm)
                for i in range(nelem)]
            )
            rdts2["h31000"] += (1.0 / 32) * np.sum([
                rbbx[i] * pxm3[i] * cpxm
                for i in range(nelem)]
            )
            t1 = np.array([np.conj(pxm[i]) * pxm for i in range(nelem)])
            t2 = np.conj(t1)
            rdts2["h11110"] += (1.0 / 16) * np.sum([
                rbxy[i] * (betaxm * (t1[i] - t2[i]) + betaym * pym2[i]
                           * cpym2 * (t2[i] + t1[i]))
                for i in range(nelem)]
            )
            t1 = np.array([np.exp(-1j * (phixm[i] - phixm))
                           for i in range(nelem)])
            t2 = np.conj(t1)
            rdts2["h11200"] += (1.0 / 32) * np.sum([
                rbxy[i] * np.exp(1j * (2 * phiym[i]))
                * (betaxm * (t1[i] - t2[i]) + 2 * betaym * (t2[i] + t1[i]))
                for i in range(nelem)]
            )
            rdts2["h40000"] += (1.0 / 64) * np.sum([
                rbbx[i] * pxm3[i] * pxm for i in range(nelem)]
            )
            rdts2["h20020"] += (1.0 / 64) * np.sum([
                rbxy[i] * (betaxm * cpxym2[i] * pxm3 - (betaxm + 4 * betaym)
                           * ppxm[i] * cpym2[i])
                for i in range(nelem)]
            )
            rdts2["h20110"] += (1.0 / 32) * np.sum([
                rbxy[i] * (betaxm * (cpxm[i] * pxm3 - ppxm[i])
                           + 2 * betaym * ppxm[i] * pym2[i] * cpym2)
                for i in range(nelem)]
            )
            rdts2["h20200"] += (1.0 / 64) * np.sum([
                rbxy[i] * (betaxm * cpxm[i] * pxm3 * pym2[i]
                           - (betaxm - 4 * betaym)
                           * ppxm[i] * pym2[i])
                for i in range(nelem)]
            )
            rdts2["h00220"] += (1.0 / 64) * np.sum([
                rbxy[i] * betaym * (pxm[i] * pym2[i] * cpxym2 + 4 * pxm[i] * cpxm
                                    - np.conj(pxm[i] * pym2) * pxm * pym2[i])
                for i in range(nelem)]
            )
            rdts2["h00310"] += (1.0 / 32) * np.sum([
                rbxy[i] * betaym * pym2[i] * (pxm[i] * cpxm - pxm * cpxm[i])
                for i in range(nelem)]
            )
            rdts2["h00400"] += (1.0 / 64) * np.sum([
                rbxy[i] * betaym * pxm[i] * cpxm * pym2[i] * pym2
                for i in range(nelem)]
            )
            # fmt: on

        if (RDTType.TUNESHIFT in rdttype) or (RDTType.ALL in rdttype):
            nelem = sum(mask_b3l)
            b3lm = b3l[mask_b3l]
            betaxm = betax[mask_b3l]
            betaym = betay[mask_b3l]
            phixm = phix[mask_b3l]
            phiym = phiy[mask_b3l]
            nux = tune[0]
            nuy = tune[1]
            # fmt: off
            b3f = [b3lm[i] * b3lm / np.pi for i in range(nelem)]
            sqbx = [np.sqrt(betaxm[i] * betaxm) for i in range(nelem)]
            bbx = [betaxm[i] * betaxm for i in range(nelem)]
            bby = [betaym[i] * betaym for i in range(nelem)]
            dphix = [abs(phixm[i] - phixm) for i in range(nelem)]
            dphiy = [abs(phiym[i] - phiym) for i in range(nelem)]
            rdts["dnux_dJx"] += np.sum([
                -b3f[i] / 16 * sqbx[i] * bbx[i]
                * (3 * np.cos(dphix[i] - np.pi * nux) / np.sin(np.pi * nux)
                   + np.cos(3 * dphix[i] - 3 * np.pi * nux)
                   / np.sin(3 * np.pi * nux))
                for i in range(nelem)]
            )
            rdts["dnux_dJy"] += np.sum([
                b3f[i] / 8 * sqbx[i] * betaym[i]
                * (2 * betaxm * np.cos(dphix[i] - np.pi * nux)
                   / np.sin(np.pi * nux)
                   - betaym * np.cos(dphix[i] + 2 * dphiy[i]
                                     - np.pi * (nux + 2 * nuy))
                   / np.sin(np.pi * (nux + 2 * nuy))
                   + betaym * np.cos(dphix[i] - 2 * dphiy[i]
                                     - np.pi * (nux - 2 * nuy))
                   / np.sin(np.pi * (nux - 2 * nuy)))
                for i in range(nelem)]
            )
            rdts["dnuy_dJy"] += np.sum([
                -b3f[i] / 16 * sqbx[i] * bby[i]
                * (4 * np.cos(dphix[i] - np.pi * nux)
                   / np.sin(np.pi * nux)
                   + np.cos(dphix[i] + 2 * dphiy[i]
                            - np.pi * (nux + 2 * nuy))
                   / np.sin(np.pi * (nux + 2 * nuy))
                   + np.cos(dphix[i] - 2 * dphiy[i]
                            - np.pi * (nux - 2 * nuy))
                   / np.sin(np.pi * (nux - 2 * nuy)))
                for i in range(nelem)]
            )
            # fmt: on
    return rdts, rdts2


def _get_rdtlist(
    idx_mag,
    beta,
    etax,
    phi,
    avemu,
    mu,
    smag,
    sall,
    pols,
    rdt_type,
    tune,
    nperiods,
    second_order,
    refpts,
):
    rdtlist = []
    rdtlist2 = []
    pf = _compute_pf(tune, nperiods)
    for ii in refpts:
        start_idx = sum(idx_mag < ii)
        beta_rot = np.roll(beta, -start_idx, axis=0)
        etax_rot = np.roll(etax, -start_idx)
        phi_rot = np.roll(phi, -start_idx, axis=0) - avemu[ii]
        if start_idx > 0:
            phi_rot[-start_idx:] += mu
        s_rot = np.roll(smag, -start_idx) - sall[ii]
        s_rot[-start_idx:] += sall[-1]
        pols_rot = np.roll(pols, -start_idx, axis=0)
        rdt, rdt2 = _computedrivingterms(
            s_rot,
            beta_rot[:, 0],
            beta_rot[:, 1],
            phi_rot[:, 0],
            phi_rot[:, 1],
            etax_rot,
            pols_rot[:, 0],
            pols_rot[:, 1],
            pols_rot[:, 2],
            pols_rot[:, 3],
            tune,
            rdt_type,
            nperiods,
            ii,
            second_order,
            pf,
        )
        rdtlist.append(rdt)
        rdtlist2.append(rdt2)
    return rdtlist, rdtlist2


def get_rdts(
    ring: Lattice,
    refpts: Refpts,
    rdt_type: Container[RDTType] | RDTType,
    second_order: bool = False,
    use_mp: bool = False,
    pool_size: int = None,
):
    """Get the lattice Resonance Driving Terms

    :py:func:`get_rdts` computes the ring RDTs based on the original implementation
    from ELEGANT. For consistency, pyAT keeps the sign convention of the AT MATLAB
    interface.

    Refs [#]_ and [#]_ were used to calculate the magnitude of first and second order
    rdts and Ref. [#]_ for the focusing rdt, however, different sign conventions are
    used along the references and in the original implementation. To overcome this
    issue, first and second order rdts are provided as a total and also separately
    so the user can redefine these conventions if needed.

    The resonance base of :py:obj:`~RDTType.GEOMETRIC2` rdts is explained in Eq. (54)
    of Ref [#]_.

    The periodicity property of the lattice is automatically taken into account in the
    rdt calculation, however the calculation of the second order contribution of
    sextupoles to the :py:obj:`~RDTType.GEOMETRIC2` and :py:obj:`~RDTType.TUNESHIFT`
    RDT types can only be derived for periodicity=1 (i.e. full ring provided).

    Parameters:
        ring: :py:class:`.Lattice` object
        refpts: Element refpts at which the RDTs are calculated
        rdt_type: Type of RDTs to be calculated.
          Possible RDT types are:

          * :py:obj:`RDTType.ALL`: all available RDTs
          * :py:obj:`RDTType.FOCUSING`: Normal quadrupole RDTs
          * :py:obj:`RDTType.COUPLING`: Linear coupling RDTs from skew quadrupoles
          * :py:obj:`RDTType.CHROMATIC`: Chromatic RDTs from sextupoles and normal
            quadrupoles
          * :py:obj:`RDTType.GEOMETRIC1`: Geometric RDTs from sextupoles
          * :py:obj:`RDTType.GEOMETRIC2`: Geometric RDTs from octupoles. The second
            order contribution of sextupoles is added when *second_order* is True
          * :py:obj:`RDTType.TUNESHIFT`: Amplitude detuning coefficients. The second
            order contribution of sextupoles is added when *second_order* is True

    Keyword Args:
        second_order (bool): Compute second order terms. Computation is significantly
          longer using this method,
        use_mp (bool):       Activate parallel calculation,
        pool_size (int):     Number of processes used for parallelization.

    Returns:
        rdts (complex): rdt data at refpts,
        rdts2 (complex): contribution from sextupole second order terms
          Available only for :py:obj:`~RDTType.GEOMETRIC2` and
          :py:obj:`~RDTType.TUNESHIFT` terms,
        rdttot (complex): total rdts.

    **rdts** is a :py:class:`record array <numpy.recarray>` with fields:

    for :py:obj:`~RDTType.FOCUSING`:
        `h20000`, `h00200`

    for :py:obj:`~RDTType.COUPLING`:
        `h10010`, `h10100`

    for :py:obj:`~RDTType.CHROMATIC`:
        `h11001`, `h00111`, `h20001`, `h00201`, `h10002`

    for :py:obj:`~RDTType.GEOMETRIC1`:
        `h21000`, `h30000`, `h10110`, `h10020`, `h10200`

    for :py:obj:`~RDTType.GEOMETRIC2`:
        `h22000`, `h11110`, `h00220`, `h31000`, `h40000`, `h20110`
        `h11200`, `h20020`, `h20200`, `h00310`, `h00400`

    for :py:obj:`~RDTType.TUNESHIFT`:
        `dnux_dJx`, `dnux_dJy`, `dnuy_dJy`

    Example:

        >>> get_rdts(ring, reftps, [RDTType.COUPLING, RDTType.CHROMATIC])

    References:
        .. [#] J.Bengtsson, SLS Note 9 / 97, March 7, 1997, with corrections per
           W.Guo (NSLS)

        .. [#] Revised to follow C.X.Wang AOP - TN - 2009 - 020 for second - order
           terms

        .. [#] A.Franchi et al. arxiv 1711.06589, PRAB 17.074001

        .. [#] Chunxi Wang and Alex Chao. Notes on Lie algebraic analysis of achromats.
           SLAC/AP-100. Jan/1995
    """
    if not isinstance(rdt_type, Container):
        rdt_type = {rdt_type}
    nperiods = ring.periodicity
    if second_order:
        assert nperiods == 1, "Second order available only for ring.periodicity=1"

    refpts = ring.uint32_refpts(refpts)
    idx_mag = ring.get_uint32_index(Multipole)
    lo, avebeta, avemu, avedisp, *_ = ring.avlinopt(refpts=All)

    sall = ring.get_s_pos(All)
    smag = sall[idx_mag]
    beta = avebeta[idx_mag]
    etax = avedisp[idx_mag, 0]
    phi = avemu[idx_mag]
    pols = [
        [
            _get_polynom(e, "PolynomA", 1) * e.Length,
            _get_polynom(e, "PolynomB", 1) * e.Length,
            _get_polynom(e, "PolynomB", 2) * e.Length,
            _get_polynom(e, "PolynomB", 3) * e.Length,
        ]
        for e in ring[idx_mag]
    ]

    mu = lo[-1].mu
    tune = mu / 2.0 / np.pi * nperiods
    fun = partial(
        _get_rdtlist,
        idx_mag,
        beta,
        etax,
        phi,
        avemu,
        mu,
        smag,
        sall,
        pols,
        rdt_type,
        tune,
        nperiods,
        second_order,
    )
    if use_mp:
        if pool_size is None:
            pool_size = min(len(refpts), multiprocessing.cpu_count())
        ctx = multiprocessing.get_context()
        refpts = np.array_split(refpts, pool_size)
        with ctx.Pool(pool_size) as pool:
            results = pool.map(fun, refpts)
            rdtlist, rdtlist2 = zip(*results)
            rdtlist = np.concatenate(rdtlist)
            rdtlist2 = np.concatenate(rdtlist2)
    else:
        rdtlist, rdtlist2 = fun(refpts)
    rdts = {}
    rdts2 = {}
    rdttot = {}
    keylist = rdtlist[0].keys()
    for k in keylist:
        val = [rdt.get(k, None) for rdt in rdtlist]
        val2 = [rdt.get(k, None) for rdt in rdtlist2]
        if val[0] is not None:
            rdts[k] = np.array(val)
            rdttot[k] = np.array(val, dtype=complex)
        if val2[0] is not None:
            rdts2[k] = np.array(val2)
            rdttot[k] += np.array(val2, dtype=complex)
    rdts2["refpts"] = rdts["refpts"]
    rdttot["refpts"] = rdts["refpts"]
    ardts = np.rec.fromarrays(rdts.values(), names=list(keylist))
    ardts2 = np.rec.fromarrays(rdts2.values(), names=list(keylist))
    ardttot = np.rec.fromarrays(rdttot.values(), names=list(keylist))
    return ardts, ardts2, ardttot


Lattice.get_rdts = get_rdts



"""Acceptance computation"""

from __future__ import annotations

__all__ = [
    "get_acceptance",
    "get_1d_acceptance",
    "get_horizontal_acceptance",
    "get_vertical_acceptance",
    "get_momentum_acceptance",
]

import multiprocessing
from typing import Sequence

import numpy as np

from .boundary import GridMode
from ..tracking import MPMode, gpu_core_count

# noinspection PyProtectedMember
from .boundary import boundary_search
from ..lattice import Lattice, Refpts, frequency_control


@frequency_control
def get_acceptance(
    ring: Lattice,
    planes,
    npoints,
    amplitudes,
    nturns: int = 1024,
    refpts: Refpts | None = None,
    dp: float | None = None,
    offset: Sequence[float] | None = None,
    bounds=None,
    grid_mode: GridMode = GridMode.RADIAL,
    use_mp: bool | MPMode = False,
    gpu_pool: list[int] | None = None,
    verbose: bool = False,
    divider: int = 2,
    shift_zero: float = 1.0e-6,
    start_method: str | None = None,
):
    # noinspection PyUnresolvedReferences
    r"""Computes the acceptance at ``repfts`` observation points

    Parameters:
        ring:           Lattice definition
        planes:         max. dimension 2, Plane(s) to scan for the acceptance.
          Allowed values are: ``'x'``, ``'xp'``, ``'y'``,
          ``'yp'``, ``'dp'``, ``'ct'``
        npoints:        (len(planes),) array: number of points in each
          dimension
        amplitudes:     (len(planes),) array: set the search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step
        nturns:         Number of turns for the tracking
        refpts:         Observation points. Default: start of the machine
        dp:             static momentum offset
        offset:         initial orbit. Default: closed orbit
        bounds:         defines the tracked range: range=bounds*amplitude.
          It can be used to select quadrants. For example, default values are:

          * :py:attr:`.GridMode.CARTESIAN`: ((-1, 1), (0, 1))
          * :py:attr:`GridMode.RADIAL/RECURSIVE <.GridMode.RADIAL>`: ((0, 1),
            (:math:`\pi`, 0))
        grid_mode:      defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp:         Flag to activate CPU or GPU multiprocessing
          (default: False)
        gpu_pool:       List of GPU id to use when use_mp is
          :py:attr:`at.tracking.MPMode.GPU`. If None specified, if gets
          first GPU.
        verbose:        Print out some information
        divider:        Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero: Epsilon offset applied on all 6 coordinates
        start_method:   Python multiprocessing start method. The default
          ``None`` uses the python default that is considered safe.
          Available parameters: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          The default for linux is ``'fork'``, the default for macOS and
          Windows is ``'spawn'``. ``'fork'`` may be used on macOS to speed up
          the calculation or to solve runtime errors, however it is
          considered unsafe.

    Returns:
        boundary:   (2,n) array: 2D acceptance
        survived:   (2,n) array: Coordinates of surviving particles
        tracked:    (2,n) array: Coordinates of tracked particles

    In case of multiple refpts, return values are lists of arrays, with one
    array per ref. point.

    Examples:

        >>> bf, sf, gf = ring.get_acceptance(planes, npoints, amplitudes)
        >>> plt.plot(*gf, ".")
        >>> plt.plot(*sf, ".")
        >>> plt.plot(*bf)
        >>> plt.show()

    .. note::

       * When``use_mp=True`` all the available CPUs will be used.
         This behavior can be changed by setting
         ``at.DConstant.patpass_poolsize`` to the desired value
       * When multiple ``refpts`` are provided particles are first
         projected to the beginning of the ring with tracking. Then,
         all particles are tracked up to ``nturns``. This allows to
         do most of the work in a single function call and allows for
         full parallelization.
    """
    kwargs = {}
    if start_method is not None:
        kwargs["start_method"] = start_method

    # For backward compatibility (use_mp can be a boolean)
    if use_mp is True:
        use_mp = MPMode.CPU

    if verbose:
        nproc = multiprocessing.cpu_count()
        print(f"\n{nproc} cpu found for acceptance calculation")
        if use_mp is MPMode.CPU:
            nprocu = nproc
            print("Multi-process acceptance calculation selected...")
            if nproc == 1:
                print("Consider use_mp=False for single core computations")
        elif use_mp is MPMode.GPU:
            nprocu = gpu_core_count(gpu_pool)
            print(f"\n{nprocu} GPU cores found")
            print("GPU acceptance calculation selected...")
            kwargs["gpu_pool"] = gpu_pool if gpu_pool is not None else [0]
        else:
            nprocu = 1
            print("Single process acceptance calculation selected...")
            if nproc > 1:
                print("Consider use_mp=True for parallelized computations")
        npts = np.atleast_1d(npoints)
        na = 2
        if len(npts) == 2:
            na = npts[1]
        npp = np.prod(npoints)
        rpp = 2 * np.ceil(np.log2(npts[0])) * np.ceil(na / nprocu)
        mpp = npp / nprocu
        if rpp > mpp:
            cond = grid_mode is GridMode.RADIAL or grid_mode is GridMode.CARTESIAN
        else:
            cond = grid_mode is GridMode.RECURSIVE
        if rpp > mpp and not cond:
            print(f"The estimated load for grid mode is {mpp}")
            print(f"The estimated load for recursive mode is {rpp}")
            print(f"{GridMode.RADIAL} or {GridMode.CARTESIAN} is recommended")
        elif rpp < mpp and not cond:
            print(f"The estimated load for grid mode is {mpp}")
            print(f"The estimated load for recursive mode is {rpp}")
            print(f"{GridMode.RECURSIVE} is recommended")

    b, s, g = boundary_search(
        ring,
        planes,
        npoints,
        amplitudes,
        nturns=nturns,
        obspt=refpts,
        dp=dp,
        offset=offset,
        bounds=bounds,
        grid_mode=grid_mode,
        use_mp=use_mp,
        verbose=verbose,
        divider=divider,
        shift_zero=shift_zero,
        **kwargs,
    )
    return b, s, g


def get_1d_acceptance(
    ring: Lattice,
    plane: str,
    resolution: float,
    amplitude: float,
    nturns: int | None = 1024,
    refpts: Refpts | None = None,
    dp: float | None = None,
    offset: Sequence[float] = None,
    grid_mode: GridMode | None = GridMode.RADIAL,
    use_mp: bool | MPMode = False,
    gpu_pool: list[int] | None = None,
    verbose: bool | None = False,
    divider: int | None = 2,
    shift_zero: float | None = 1.0e-6,
    start_method: str | None = None,
):
    r"""Computes the 1D acceptance at ``refpts`` observation points

    See :py:func:`get_acceptance`

    Parameters:
        ring:           Lattice definition
        plane:          Plane to scan for the acceptance.
          Allowed values are: ``'x'``, ``'xp'``, ``'y'``, ``'yp'``, ``'dp'``,
          ``'ct'``
        resolution:     Minimum distance between 2 grid points
        amplitude:      Search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step
        nturns:         Number of turns for the tracking
        refpts:         Observation points. Default: start of the machine
        dp:             static momentum offset
        offset:         initial orbit. Default: closed orbit
        grid_mode:      defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp:         Flag to activate CPU or GPU multiprocessing
          (default: False). In case multiprocessing is not enabled,
          ``grid_mode`` is forced to :py:attr:`.GridMode.RECURSIVE`
          (most efficient in single core)
        gpu_pool:       List of GPU id to use when use_mp is
          :py:attr:`at.tracking.MPMode.GPU`. If None specified, if gets
          first GPU.
        verbose:        Print out some information
        divider:        Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero: Epsilon offset applied on all 6 coordinates
        start_method:   Python multiprocessing start method. The default
          ``None`` uses the python default that is considered safe.
          Available parameters: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          The default for linux is ``'fork'``, the default for macOS and
          Windows is ``'spawn'``. ``'fork'`` may be used on macOS to speed up
          the calculation or to solve runtime errors, however it is considered
          unsafe.

    Returns:
        boundary:   (len(refpts),2) array: 1D acceptance
        survived:   (n,) array: Coordinates of surviving particles
        tracked:    (n,) array: Coordinates of tracked particles

    In case of multiple ``tracked`` and ``survived`` are lists of arrays,
    with one array per ref. point.

    .. note::

       * When``use_mp=True`` all the available CPUs will be used.
         This behavior can be changed by setting
         ``at.DConstant.patpass_poolsize`` to the desired value
       * When multiple ``refpts`` are provided particles are first
         projected to the beginning of the ring with tracking. Then,
         all particles are tracked up to ``nturns``. This allows to
         do most of the work in a single function call and allows for
         full parallelization.
    """
    if not use_mp:
        if verbose:
            print("No parallel calculation selected, force to GridMode.RECURSIVE")
        grid_mode = GridMode.RECURSIVE
    assert len(np.atleast_1d(plane)) == 1, "1D acceptance: single plane required"
    assert np.isscalar(resolution), "1D acceptance: scalar args required"
    assert np.isscalar(amplitude), "1D acceptance: scalar args required"
    npoint = np.ceil(amplitude / resolution)
    if grid_mode is not GridMode.RECURSIVE:
        assert npoint > 1, (
            "Grid has only one point: increase amplitude or reduce resolution"
        )
    b, s, g = get_acceptance(
        ring,
        plane,
        npoint,
        amplitude,
        nturns=nturns,
        dp=dp,
        refpts=refpts,
        grid_mode=grid_mode,
        use_mp=use_mp,
        gpu_pool=gpu_pool,
        verbose=verbose,
        start_method=start_method,
        divider=divider,
        shift_zero=shift_zero,
        offset=offset,
    )
    return np.squeeze(b), s, g


def get_horizontal_acceptance(
    ring: Lattice, resolution: float, amplitude: float, *args, **kwargs
):
    r"""Computes the 1D horizontal acceptance at ``refpts`` observation points

    See :py:func:`get_acceptance`

    Parameters:
        ring:           Lattice definition
        resolution:     Minimum distance between 2 grid points
        amplitude:      Search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step

    Keyword Args:
        nturns:         Number of turns for the tracking
        refpts:         Observation points. Default: start of the machine
        dp:             static momentum offset
        offset:         initial orbit. Default: closed orbit
        grid_mode:      defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp:         Use python multiprocessing (:py:func:`.patpass`,
          default use :py:func:`.lattice_pass`). In case multiprocessing
          is not enabled, ``grid_mode`` is forced to
          :py:attr:`.GridMode.RECURSIVE` (most efficient in single core)
        gpu_pool:       List of GPU id to use when use_mp is
          :py:attr:`at.tracking.MPMode.GPU`. If None specified, if gets
          first GPU.
        verbose:        Print out some information
        divider:        Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero: Epsilon offset applied on all 6 coordinates
        start_method:   Python multiprocessing start method. The default
          ``None`` uses the python default that is considered safe.
          Available parameters: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          The default for linux is ``'fork'``, the default for macOS and
          Windows is ``'spawn'``. ``'fork'`` may be used on macOS to speed up
          the calculation or to solve runtime errors, however it is considered
          unsafe.

    Returns:
        boundary:   (len(refpts),2) array: 1D acceptance
        survived:   (n,) array: Coordinates of surviving particles
        tracked:    (n,) array: Coordinates of tracked particles

    In case of multiple ``tracked`` and ``survived`` are lists of arrays,
    with one array per ref. point.

    .. note::

       * When``use_mp=True`` all the available CPUs will be used.
         This behavior can be changed by setting
         ``at.DConstant.patpass_poolsize`` to the desired value
       * When multiple ``refpts`` are provided particles are first
         projected to the beginning of the ring with tracking. Then,
         all particles are tracked up to ``nturns``. This allows to
         do most of the work in a single function call and allows for
         full parallelization.
    """
    return get_1d_acceptance(ring, "x", resolution, amplitude, *args, **kwargs)


def get_vertical_acceptance(
    ring: Lattice, resolution: float, amplitude: float, *args, **kwargs
):
    r"""Computes the 1D vertical acceptance at refpts observation points

    See :py:func:`get_acceptance`

    Parameters:
        ring:           Lattice definition
        resolution:     Minimum distance between 2 grid points
        amplitude:      Search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step

    Keyword Args:
        nturns:         Number of turns for the tracking
        refpts:         Observation points. Default: start of the machine
        dp:             static momentum offset
        offset:         initial orbit. Default: closed orbit
        grid_mode:      defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp:         Use python multiprocessing (:py:func:`.patpass`,
          default use :py:func:`.lattice_pass`). In case multiprocessing
          is not enabled, ``grid_mode`` is forced to
          :py:attr:`.GridMode.RECURSIVE` (most efficient in single core)
        gpu_pool:       List of GPU id to use when use_mp is
          :py:attr:`at.tracking.MPMode.GPU`. If None specified, if gets
          first GPU.
        verbose:        Print out some information
        divider:        Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero: Epsilon offset applied on all 6 coordinates
        start_method:   Python multiprocessing start method. The default
          ``None`` uses the python default that is considered safe.
          Available parameters: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          The default for linux is ``'fork'``, the default for macOS and
          Windows is ``'spawn'``. ``'fork'`` may be used on macOS to speed up
          the calculation or to solve runtime errors, however it is considered
          unsafe.

    Returns:
        boundary:   (len(refpts),2) array: 1D acceptance
        survived:   (n,) array: Coordinates of surviving particles
        tracked:    (n,) array: Coordinates of tracked particles

    In case of multiple ``tracked`` and ``survived`` are lists of arrays,
    with one array per ref. point.

    .. note::

       * When``use_mp=True`` all the available CPUs will be used.
         This behavior can be changed by setting
         ``at.DConstant.patpass_poolsize`` to the desired value
       * When multiple ``refpts`` are provided particles are first
         projected to the beginning of the ring with tracking. Then,
         all particles are tracked up to ``nturns``. This allows to
         do most of the work in a single function call and allows for
         full parallelization.
    """
    return get_1d_acceptance(ring, "y", resolution, amplitude, *args, **kwargs)


def get_momentum_acceptance(
    ring: Lattice, resolution: float, amplitude: float, *args, **kwargs
):
    r"""Computes the 1D momentum acceptance at refpts observation points

    See :py:func:`get_acceptance`

    Parameters:
        ring:           Lattice definition
        resolution:     Minimum distance between 2 grid points
        amplitude:      Search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step

    Keyword Args:
        nturns:         Number of turns for the tracking
        refpts:         Observation points. Default: start of the machine
        dp:             static momentum offset
        offset:         initial orbit. Default: closed orbit
        grid_mode:      defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp:         Use python multiprocessing (:py:func:`.patpass`,
          default use :py:func:`.lattice_pass`). In case multiprocessing is
          not enabled, ``grid_mode`` is forced to
          :py:attr:`.GridMode.RECURSIVE` (most efficient in single core)
        gpu_pool:       List of GPU id to use when use_mp is
          :py:attr:`at.tracking.MPMode.GPU`. If None specified, if gets
          first GPU.
        verbose:        Print out some information
        divider:        Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero: Epsilon offset applied on all 6 coordinates
        start_method:   Python multiprocessing start method. The default
          ``None`` uses the python default that is considered safe.
          Available parameters: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          The default for linux is ``'fork'``, the default for macOS and
          Windows is ``'spawn'``. ``'fork'`` may be used on macOS to speed up
          the calculation or to solve runtime errors, however it is considered
          unsafe.

    Returns:
        boundary:   (len(refpts),2) array: 1D acceptance
        survived:   (n,) array: Coordinates of surviving particles
        tracked:    (n,) array: Coordinates of tracked particles

    In case of multiple ``tracked`` and ``survived`` are lists of arrays,
    with one array per ref. point.

    .. note::

       * When``use_mp=True`` all the available CPUs will be used.
         This behavior can be changed by setting
         ``at.DConstant.patpass_poolsize`` to the desired value
       * When multiple ``refpts`` are provided particles are first
         projected to the beginning of the ring with tracking. Then,
         all particles are tracked up to ``nturns``. This allows to
         do most of the work in a single function call and allows for
         full parallelization.
    """
    return get_1d_acceptance(ring, "dp", resolution, amplitude, *args, **kwargs)


Lattice.get_acceptance = get_acceptance
Lattice.get_horizontal_acceptance = get_horizontal_acceptance
Lattice.get_vertical_acceptance = get_vertical_acceptance
Lattice.get_momentum_acceptance = get_momentum_acceptance



from __future__ import annotations
from ..lattice import Lattice
from ..latticetools import ResponseMatrix
from typing import Optional
import matplotlib.pyplot as plt
from matplotlib.axes import Axes


def plot_norm(resp: ResponseMatrix, ax: Optional[tuple[Axes, Axes]] = None) -> None:
    r"""Plot the norm of the lines and columns of the weighted response matrix

    For a stable solution, the norms should have the same order of magnitude.
    If not, the weights of observables and variables should be adjusted.

    Args:
        resp:           Response matrix object
        ax:             tuple of :py:class:`~.matplotlib.axes.Axes`. If given,
          plots will be drawn in these axes.
    """
    obs, var = resp.check_norm()
    if ax is None:
        fig, (ax1, ax2) = plt.subplots(nrows=2, layout="constrained")
    else:
        ax1, ax2 = ax[:2]
    ax1.bar(range(len(obs)), obs)
    ax1.set_title("Norm of weighted observables")
    ax1.set_xlabel("Observable #")
    ax2.bar(range(len(var)), var)
    ax2.set_title("Norm of weighted variables")
    ax2.set_xlabel("Variable #")


def plot_singular_values(
    resp: ResponseMatrix, ax: Axes = None, logscale: bool = True
) -> None:
    r"""Plot the singular values of a response matrix

    Args:
        resp:           Response matrix object
        logscale:       If :py:obj:`True`, use log scale
        ax:             If given, plots will be drawn in these axes.
    """
    if resp.singular_values is None:
        resp.solve()
    singvals = resp.singular_values
    if ax is None:
        fig, ax = plt.subplots()
    ax.bar(range(len(singvals)), singvals)
    if logscale:
        ax.set_yscale("log")
    ax.set_title("Singular values")


def plot_obs_analysis(
    resp: ResponseMatrix, lattice: Lattice, ax: Axes = None, logscale: bool = True
) -> None:
    """Plot the decomposition of an error vector on the basis of singular
    vectors

    Args:
        resp:           Response matrix object
        lattice:        Lattice description. The response matrix observables
          will be evaluated for this :py:class:`.Lattice` and the deviation
          from   target will be decomposed on the basis of singular vectors,
        logscale:       If :py:obj:`True`, use log scale
        ax:             If given, plots will be drawn in these axes.
    """
    if resp.singular_values is None:
        resp.solve()
    obs = resp.observables
    # noinspection PyProtectedMember
    obs.evaluate(lattice, **resp._eval_args)
    corr = resp._uh @ obs.flat_deviations
    if ax is None:
        fig, ax = plt.subplots()
    ax.bar(range(len(corr)), corr)
    if logscale:
        ax.set_yscale("log")
    ax.set_title("SVD decomposition")
    ax.set_xlabel("Singular vector #")


def plot_var_analysis(
    resp: ResponseMatrix, lattice: Lattice, ax: Axes = None, logscale: bool = False
) -> None:
    """Plot the decomposition of a correction vector on the basis of singular
    vectors

    Args:
        resp:           Response matrix object
        lattice:        Lattice description. The variables will be evaluated
          for this :py:class:`.Lattice` and will be decomposed on the basis
          of singular vectors,
        logscale:       If :py:obj:`True`, use log scale
        ax:             If given, plots will be drawn in these axes.
    """
    if resp.singular_values is None:
        resp.solve()
    var = resp.variables
    if ax is None:
        fig, ax = plt.subplots()
    corr = (resp._v * resp.singular_values).T @ var.get(lattice)
    ax.bar(range(len(corr)), corr)
    if logscale:
        ax.set_yscale("log")
    ax.set_title("SVD decomposition")
    ax.set_xlabel("Singular vector #")


ResponseMatrix.plot_norm = plot_norm
ResponseMatrix.plot_singular_values = plot_singular_values
ResponseMatrix.plot_obs_analysis = plot_obs_analysis
ResponseMatrix.plot_var_analysis = plot_var_analysis



"""Revolution frequency, momentum compaction factor, slip factor"""
from __future__ import annotations

__all__ = ["get_mcf", "get_slip_factor", "get_revolution_frequency"]

import numpy as np

from .orbit import find_orbit4
from ..constants import clight
from ..lattice import DConstant, Lattice, check_6d, get_s_pos
from ..tracking import internal_lpass


@check_6d(False)
def get_mcf(
    ring: Lattice,
    dp: float = 0.0,
    keep_lattice: bool = False,
    fit_order: int = 1,
    n_step: int = 2,
    **kwargs,
) -> float:
    r"""Compute the momentum compaction factor :math:`\alpha`

    Parameters:
        ring:           Lattice description (:pycode:`ring.is_6d` must be
          :py:obj:`False`)
        dp:             Momentum deviation
        keep_lattice:   Assume no lattice change since the previous tracking.
        fit_order:      Maximum momentum compaction factor order to be fitted.
          Default to 1, corresponding to the first-order momentum compaction factor.
        n_step:         Number of different calculated momentum deviations to be fitted
          with a polynomial.
          Default to 2.

    Keyword Args:
        DPStep (float): Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`

    Returns:
        mcf (float/array):    Momentum compaction factor :math:`\alpha`
          up to the order *fit_order*. Returns a float if *fit_order==1* otherwise
          returns an array.
    """
    if n_step < 2 * fit_order:
        raise ValueError(
            "Low number of steps, it is advised to have n_step >= 2*fit_order"
            + " for a better fit."
        )
    dp_step = kwargs.pop("DPStep", DConstant.DPStep)
    dp_samples = np.linspace(-dp_step / 2, dp_step / 2, n_step)
    fp_i = tuple(
        find_orbit4(ring, dp=dp + dp_i, keep_lattice=keep_lattice)[0]
        for dp_i in dp_samples
    )
    fp = np.stack(fp_i, axis=0).T  # generate a Fortran contiguous array
    b = np.squeeze(internal_lpass(ring, fp, keep_lattice=True), axis=(2, 3))
    ring_length = get_s_pos(ring, len(ring))
    p = np.polynomial.Polynomial.fit(b[4], b[5], deg=fit_order).convert().coef
    alphac = p[1:] / ring_length[0]
    return alphac[0] if len(alphac) < 2 else alphac


def get_slip_factor(ring: Lattice, **kwargs) -> float:
    r"""Compute the slip factor :math:`\eta=1/\gamma^2-\alpha`

    Parameters:
        ring:           Lattice description (:pycode:`ring.is_6d` must be
          :py:obj:`False`)

    Keyword Args:
        dp (float):     Momentum deviation
        DPStep (float): Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`

    Returns:
        eta (float):    Slip factor :math:`\eta`
    """
    gamma = ring.gamma
    etac = 1.0 / gamma / gamma - get_mcf(ring, **kwargs)
    return etac


def get_revolution_frequency(
    ring: Lattice, dp: float = None, dct: float = None, df: float = None
) -> float:
    """Compute the revolution frequency of the full ring [Hz]

    Parameters:
        ring:       Lattice description
        dp:         Momentum deviation. Defaults to :py:obj:`None`
        dct:        Path lengthening. Defaults to :py:obj:`None`
        df:         Deviation of RF frequency. Defaults to :py:obj:`None`

    Returns:
        frev:       Revolution frequency [Hz]
    """
    lcell = ring.cell_length
    cell_frev = ring.beta * clight / lcell
    if dct is not None:
        cell_frev *= lcell / (lcell + dct)
    elif dp is not None:
        # Find the path lengthening for dp
        rnorad = ring.disable_6d(copy=True) if ring.is_6d else ring
        orbit = internal_lpass(rnorad, rnorad.find_orbit4(dp=dp)[0])
        dct = np.squeeze(orbit)[5]
        cell_frev *= lcell / (lcell + dct)
    elif df is not None:
        cell_frev += df / ring.cell_harmnumber
    return cell_frev / ring.periodicity


Lattice.mcf = property(get_mcf, doc="Momentum compaction factor")
Lattice.slip_factor = property(get_slip_factor, doc="Slip factor")
Lattice.get_mcf = get_mcf
Lattice.get_slip_factor = get_slip_factor
Lattice.get_revolution_frequency = get_revolution_frequency



"""Defines classes for modifying a lattice and observing its parameters."""

from .observables import *
from .rdt_observable import *
from .observablelist import *
# from .matching import *
from .response_matrix import *



import numpy
from .lattice_object import Lattice
from .elements import Element
from .utils import Refpts, BoolRefpts, Uint32Refpts
from .utils import checkattr, get_uint32_index, get_bool_index
from typing import Sequence

__all__ = ['get_cells', 'get_refpts']


# noinspection PyIncorrectDocstring
def get_cells(ring: Sequence[Element], refpts: Refpts, *args,
              regex=False) -> BoolRefpts:
    # noinspection PyShadowingNames
    r"""
    get_cells(ring, filtfunc) -> BoolRefpts
    get_cells(ring, element_type) -> BoolRefpts
    get_cells(ring, attrname) -> BoolRefpts
    get_cells(ring, attrname, attrvalue) -> BoolRefpts
    Returns a bool array of element indices, selecting ring elements.

    Deprecated: :pycode:`get_cells(ring, refpts)` is
    :pycode:`ring.bool_refpts(refpts)` except for :py:obj:`str` arguments:
    :pycode:`get_cells(ring, attrname [, attrvalue])` is
    :pycode:`ring.bool_refpts(checkattr(strkey [, attrvalue]))`

    Parameters:
        ring (Sequence[Element]):       Lattice description
        filtfunc (ElementFilter):       Filter function. Selects
          :py:class:`.Element`\ s satisfying the filter function
        element_type (Type[Element]):   Element type
        attrname (str):                 Attribute name
        attrvalue (Any):                Attribute value. If absent, select the
          presence of an *attrname* attribute. If present, select
          :py:class:`.Element`\ s with :pycode:`attrname == attrvalue`.
        regex: Use regular expression for refpts string matching;
            Default: False (Unix shell-style wildcards)

    Returns:
        bool_refs (BoolRefpts):  numpy Array of :py:obj:`bool` with length
          len(ring)+1

    Examples:

        >>> refpts = get_cells(ring, 'Frequency')

        Returns a numpy array of booleans where all elements having a
        :pycode:`Frequency` attribute :py:obj:`True`

        >>> refpts = get_cells(ring, 'K', 0.0)

        Returns a numpy array of booleans where all elements having a
        :pycode:`K` attribute equal to 0.0 are :py:obj:`True`

    See also:
        :py:meth:`.Lattice.bool_refpts`, :py:meth:`.Lattice.uint32_refpts`
    """
    if isinstance(refpts, str):
        refpts = checkattr(refpts, *args)
    return get_bool_index(ring, refpts, regex=regex)


# noinspection PyUnusedLocal,PyIncorrectDocstring
def get_refpts(ring: Sequence[Element], refpts: Refpts,
               regex=False) -> Uint32Refpts:
    r"""Return a :py:obj:`~numpy.uint32` array of element indices selecting
    ring elements.

    Deprecated: :pycode:`get_elements(ring, refpts)` is
    :pycode:`ring.uint32_refpts(refpts)`

    Parameters:
        ring:           Lattice description
        refpts:         Element selection key.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        regex: Use regular expression for refpts string matching;
            Default: False (Unix shell-style wildcards)

    Returns:
        uint32_refs (Uint32Refs):    :py:obj:`~numpy.uint32` numpy array as
          long as the number of refpts

    See also:
        :py:meth:`.Lattice.uint32_refpts`, :py:meth:`.Lattice.bool_refpts`
    """
    return get_uint32_index(ring, refpts, regex=regex)


def rotate_elem(elem: Element, tilt: float = 0.0, pitch: float = 0.0,
                yaw: float = 0.0, relative: bool = False) -> None:
    r"""Set the tilt, pitch, and yaw angle of an :py:class:`.Element`.
    The tilt is a rotation around the *s*-axis, the pitch is a
    rotation around the *x*-axis, and the yaw is a rotation around
    the *y*-axis.

    A positive angle represents a clockwise rotation when
    looking in the direction of the rotation axis.

    The transformations are not all commutative, the pitch and yaw
    are applied first, and the tilt is always the last transformation
    applied. The element is rotated around its mid-point.

    If *relative* is :py:obj:`True`, the previous angle and shifts
    are rebuilt from the *R* and *T* matrix and incremented by the
    input arguments.

    The shift is always conserved regardless of the value of *relative*.

    The transformations are applied by changing the particle coordinates
    at the entrance of the element and restoring them at the end. Following
    the small angles approximation, the longitudinal shift of the particle
    coordinates is neglected and the element length is unchanged.

    Parameters:
        elem:           Element to be tilted
        tilt:           Tilt angle [rad]
        pitch:          Pitch angle [rad]
        yaw:            Yaw angle [rad]
        relative:       If :py:obj:`True`, the rotation is added to the
          previous one
    """
    # noinspection PyShadowingNames
    def _get_rm_tv(le, tilt, pitch, yaw):
        tilt = numpy.around(tilt, decimals=15)
        pitch = numpy.around(pitch, decimals=15)
        yaw = numpy.around(yaw, decimals=15)
        ct, st = numpy.cos(tilt), numpy.sin(tilt)
        ap, ay = 0.5*le*numpy.tan(pitch), 0.5*le*numpy.tan(yaw)
        rr1 = numpy.asfortranarray(numpy.diag([ct, ct, ct, ct, 1.0, 1.0]))
        rr1[0, 2] = st
        rr1[1, 3] = st
        rr1[2, 0] = -st
        rr1[3, 1] = -st
        rr2 = rr1.T
        t1 = numpy.array([ay, numpy.sin(-yaw), -ap, numpy.sin(pitch), 0, 0])
        t2 = numpy.array([ay, numpy.sin(yaw), -ap, numpy.sin(-pitch), 0, 0])
        rt1 = numpy.eye(6, order='F')
        rt1[1, 4] = t1[1]
        rt1[3, 4] = t1[3]
        rt2 = numpy.eye(6, order='F')
        rt2[1, 4] = t2[1]
        rt2[3, 4] = t2[3]
        return rr1 @ rt1, rt2 @ rr2, t1, t2

    tilt0 = 0.0
    pitch0 = 0.0
    yaw0 = 0.0
    t10 = numpy.zeros(6)
    t20 = numpy.zeros(6)
    if hasattr(elem, 'R1') and hasattr(elem, 'R2'):
        rr10 = numpy.eye(6, order='F')
        rr10[:4, :4] = elem.R1[:4, :4]
        rt10 = rr10.T @ elem.R1
        tilt0 = numpy.arctan2(rr10[0, 2], rr10[0, 0])
        yaw0 = numpy.arcsin(-rt10[1, 4])
        pitch0 = numpy.arcsin(rt10[3, 4])
        _, _, t10, t20 = _get_rm_tv(elem.Length, tilt0, pitch0, yaw0)
    if hasattr(elem, 'T1') and hasattr(elem, 'T2'):
        t10 = elem.T1-t10
        t20 = elem.T2-t20
    if relative:
        tilt += tilt0
        pitch += pitch0
        yaw += yaw0

    r1, r2, t1, t2 = _get_rm_tv(elem.Length, tilt, pitch, yaw)
    elem.R1 = r1
    elem.R2 = r2
    elem.T1 = t1+t10
    elem.T2 = t2+t20


Lattice.uint32_refpts = get_uint32_index
Lattice.bool_refpts = get_bool_index
Lattice.get_cells = get_cells
Lattice.get_refpts = get_refpts



"""RPN interpreter. From Onel Harrison:
https://onelharrison.medium.com/watch-building-a-reverse-polish-notation-rpn-evaluator-in-python-75b1c910fab6
"""
from __future__ import annotations

__all__ = ["Rpn"]

import operator as op
from math import sqrt, sin, cos, pi, pow
from typing import Any, Union

Number = Union[int, float, str]


# noinspection PyUnusedLocal
def _pop(x):
    return []


def _swap(x, y):
    return [y, x]


supported_operators = {
    "+": (op.add, 2),
    "-": (op.sub, 2),
    "*": (op.mul, 2),
    "/": (op.truediv, 2),
    "sqrt": (sqrt, 1),
    "pi": (pi, 0),
    "sin": (sin, 1),
    "cos": (cos, 1),
    "pow": (pow, 2),
    "pop": (_pop, 1),
    "swap": (_swap, 2),
}


class Rpn:

    def __init__(self):
        self.stack = []
        self.store = False
        self.operators = supported_operators.copy()

    def _mpop(self, n: int = 1) -> list[Any]:
        """Pops and returns `n` items from the stack."""
        try:
            return [self.stack.pop() for _ in range(n)]
        except IndexError:
            raise SyntaxError("RPN: Malformed expression: empty stack") from None

    @staticmethod
    def _to_num(x: Any) -> Number:
        """Converts a value to its appropriate numeric type"""
        try:
            n = float(x)
        except ValueError:
            return x
        else:
            return int(n) if n.is_integer() else n

    def _consume_token(self, token: str) -> None:
        """Consumes a token given the current stack and returns the updated stack"""
        if token == "sto":
            self.store = True
            return
        elif self.store:
            self.operators[token] = (self.stack[-1], 0)
            self.store = False
            return
        try:
            oper, narg = self.operators[token]
        except KeyError:
            result = self._to_num(token)
        else:
            if narg == 0:
                result = oper
            else:
                result = oper(*reversed(self._mpop(narg)))

        if isinstance(result, list):
            self.stack.extend(result)
        else:
            self.stack.append(result)

    def input(self, expr: str) -> None:
        for token in expr.split():
            self._consume_token(token)

    def evaluate(self, expr: str) -> Number:
        """Evaluate an RPN expression and return the result"""
        for token in expr.split():
            self._consume_token(token)

        result, = self._mpop(1)
        if self.stack:
            raise SyntaxError("RPN: Found extra tokens")
        return result



from __future__ import annotations

__all__ = ["Exporter"]

import sys
from collections.abc import Sequence, Generator

from .file_input import ElementDescr
from ..lattice import Lattice, elements as elt


class Exporter:
    delimiter: str = ";"
    continuation: str = ""
    label_fmt = str.maketrans("*/+-", "..__")  # Not allowed in destination format
    bool_fmt = None
    use_line = True

    def __init__(self, ring: Lattice, **kwargs):
        def store_elem(store, elem: ElementDescr):
            """Store an element in the selected dictionary"""

            def name_gen(name: str, max: int = 10000):
                yield name
                for i in range(1, max):
                    yield ".".join((name, str(i)))

            def check(name: str) -> tuple[bool, bool]:
                for st in self.all_stores:
                    if name in st:
                        if st is store:
                            if elem == st[name]:
                                return True, False
                            else:
                                return False, False
                        else:
                            return False, False
                return True, True

            for nm in name_gen(elem.name.translate(self.label_fmt)):
                valid, okstore = check(nm)
                if valid:
                    if okstore:
                        store[nm] = elem
                    return nm

            raise NameError(f"Cannot store {elem.name}")

        def scan(ring: Lattice) -> Generator[tuple[str, ElementDescr], None, None]:
            """Run through the lattice and store the converted elements"""
            end = 0.0
            for atelem in ring:
                attyp = type(atelem)
                elms = self.generate_madelems(attyp, vars(atelem).copy())
                if isinstance(elms, ElementDescr):
                    elms = (elms,)
                for elem in elms:
                    store = self.store.get(attyp, self.anystore)
                    length = elem.get("L", 0.0)
                    loc = end + 0.5 * length
                    end += length
                    if store is not None:
                        name = store_elem(store, elem)
                        yield name, loc

        use_line = kwargs.pop("use_line", self.use_line)
        self.seqname = kwargs.pop("use", ring.name if ring.name else "RING")
        self.anystore = {}
        self.store = {
            elt.Drift: {} if use_line else None,
            elt.Quadrupole: {},
            elt.Sextupole: {},
            elt.Dipole: {},
        }
        self.all_stores = [st for st in self.store.values() if st is not None] + [
            self.anystore
        ]
        ElementDescr.bool_fmt = self.bool_fmt
        self.seq = list(scan(ring))
        self.length = ring.cell_length
        self.in_file = getattr(ring, "in_file", "<unknown>")
        self.in_use = getattr(ring, "use", "<unknown")
        self.energy = ring.energy
        self.particle = ring.particle
        self.is_6d = ring.is_6d
        # self.anystore[self.seqname] = None

    def generate_madelems(
        self, eltype: type[elt.Element], elemdict: dict
    ) -> ElementDescr | list[ElementDescr]:
        pass

    def print_beam(self, file):
        pass

    def print_elems(self, store: dict, file) -> None:
        print(file=file)
        for elname, el in store.items():
            if el is not None:
                print(f"{elname.ljust(10)}: {str(el)}{self.delimiter}", file=file)

    def print_sequence(self, file) -> None:
        line = f"\n{self.seqname.ljust(10)}: SEQUENCE, L={self.length}{self.delimiter}"
        print(f"\n{line}{self.delimiter}", file=file)
        for elname, at in self.seq:
            print(f"  {elname.ljust(10)}, AT={at}{self.delimiter}", file=file)
        print(f"ENDSEQUENCE{self.delimiter}", file=file)

    def print_line(self, file) -> None:
        def elnames(subseq: Sequence):
            return ", ".join(ename.ljust(8) for ename, at in subseq)

        nl = int((len(self.seq) - 1) / 10)
        print(f"\n{self.seqname.ljust(10)}: LINE=( {self.continuation}", file=file)
        for lnum in range(nl):
            print(
                f"  {elnames(self.seq[10*lnum:10*(lnum+1)])}, {self.continuation}",
                file=file,
            )
        print(f"  {elnames(self.seq[10*nl:])}){self.delimiter}\n", file=file)

    def export(self, filename: str | None = None) -> None:
        def do_export(file):
            print(
                f"! Converted by PyAT from in_file={self.in_file}, use={self.in_use!r}\n",
                file=file,
            )
            self.print_beam(file)
            for store in self.all_stores:
                if store is not None:
                    self.print_elems(store, file)

            if self.store[elt.Drift] is None:
                self.print_sequence(file)
            else:
                self.print_line(file)

        if filename is None:
            do_export(sys.stdout)
        else:
            with open(filename, "w") as mfile:
                do_export(mfile)



"""Stub file for the 'diffmatrix' extension"""

from __future__ import annotations

import numpy as np

from . import Orbit
from ..lattice import Element

def FDW(element: Element, orbit: Orbit, energy: float | None = None) -> np.ndarray: ...
